; Version of 17 November 2008

pro afrl_bcd_overlap, infiles, $

    ; overlap analysis keyword parameters
    MASK_FILES=mask_files, BAD_BITS=bad_bits, BAD_BIT_CODE=bad_bit_code, $
    MIN_OVERLAP=min_overlap, MAX_FLUX=max_flux, FLUX_WEIGHTING=flux_weighting, $
    WEIGHT_OVERLAPS=weight_overlaps, WEIGHT_THRESH=weight_thresh, $
    OVERLAPS_PER_BCD=overlaps_per_bcd, REDUCE=reduce, SOURCE_FILE=source_file, $

    ; environment-saving keywords
    SAVE_FILE=save_file, RESTORE_FILE=restore_file, $

    ; offset solution keyword parameters
    OUTFILES=outfiles, OUT_DIR=out_dir, $
    N_ITERATIONS=n_iterations, STIFFNESS=stiffness, $
    OUTLIER_THRESH=outlier_thresh, WEIGHTS=weights, $
    AOR_DEWEIGHT=aor_deweight, $

    ; group offset correction keywords
    GROUP_TAGS=group_tags, GROUP_AOR=group_aor, GROUP_STIFFNESS=group_stiffness, $ 
    GROUP_ANCHORS=group_anchors, GROUP_FLOATS=group_floats, $
    GROUP_OUTLIER_THRESH=group_outlier_thresh, GROUP_EQUALIZE=group_equalize, $
    INTRA_GROUP=intra_group, GROUP_ONEPASS=group_onepass, $
    GROUP_DEWEIGHT=group_deweight, $

    ; Diagnostics
    DIAGNOSTICS=diagnostics

;+ 
; NAME:
;    AFRL_BCD_OVERLAP
;
; PURPOSE:
;
;    Match background levels for an ensemble of BCDs by assigning to each a
;    scalar offset, calculated by a least-squares minimization of the 
;    overlap differences between BCDs.
;
;    This procedure is optimized for very large data sets but will work 
;    efficiently on almost any set of BCDs. There is a hardwired upper
;    limit of 10^6 BCDs; typical systems should be able to handle several
;    hundred thousand BCDs, depending in part on the distribution of the
;    BCDs on the sky. 
;
;    Note that all BCDs must have the same array dimensions and 
;    pixel size ("plate scale").
;
;    This version will optionally perform an initial overlap correction of
;    groups of BCDs (e.g. AORs) to match varying overall background levels.
;    See the group correction keywords below.
;
; BASIC USAGE:
;
;    afrl_bcd_overlap, infiles, OUTFILES=outfiles, MASK_FILES=mask_files
;
;    infiles = list (vector) of input BCD file names 
;    outfiles = list of output BCD names to write (not required)
;    mask_files = list of bit mask (bmask) files corresponding to infiles BCDs
;
;    If MASK_FILES are not specified, then a routine called return_valid_pixels
;    must be defined, which takes an image array as input parameter and produces
;    output byte mask parameter (through some empirical criteria):
;        return_valid_pixels, im, outmask   ;  outmask=1B for "use" and 0B for "don't use"
;
;    See STRATEGIES section below for recommendations for different situations.
;
;
; INPUTS:
;    infiles: vector of filenames of BCDs to overlap-correct
;
; OUTPUTS:
;    None specifically.
;    Set OUTFILES keyword to specify output BCD file names (optional),
;        or set OUT_DIR to specify output directory (with output BCD file names
;        the same as the input).
;        OUT_DIR will not overwrite existing files.
;
; ENVIRONMENT KEYWORDS
;
;    The process can be divided into two parts: the time-consuming overlap analysis 
;        and the much faster solution of the resulting matrices. To try out different
;        solution options for a given BCD set, use save_file keyword to save the 
;        overlap analysis, and restore_file to pick up after the overlap analysis
;        and apply the specified solution options.
;        > aflr_bcd_overlap, infiles, [overlap analysis options], save_file='overlap.sav'
;        > aflr_bcd_overlap, 1, [solution options], restore_file='overlap.sav'
;    SAVE_FILE : Specify file name for IDL save. This will be performed after
;        the overlap analysis, and the routine will exit. Specify whichever
;        "overlap analysis" keywords, and leave the others unset.
;        Don't set any "offset solution" keywords unless you know you aren't
;        going to try different values, as setting them in the save file will
;        override any values set when you run the routine with restore_file.
;    RESTORE_FILE : Restore the saved environment, and jump directly to the offset 
;        solution. Set whichever offset solution keywords as desired.
;        ("infiles" parameter is ignored but needs to be set to some nonzero value to
;        keep the routine from exiting.)
;
; OVERLAP ANALYSIS KEYWORDS:
;
;    MASK_FILES : if set should be a vector of filenames of bit mask (bmask) files
;        corresponding 1-to-1 with the input BCDs. If not set need to 
;        have available an IDL program called return_valid_pixels which
;        takes as input an image array and returns a corresponding byte mask:
;            return_valid_pixels, image, mask ; mask=0B for don't use, 1B for use.
;    BAD_BITS : Array of mask bits to apply in masking images (if using MASK_FILES), 
;        default = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
;    BAD_BIT_CODE = Short int representation of bad bit mask. 
;        i.e. bad_bit_code = fix(total(2^bad_bits)), equivalent default = 32764
;        Setting this overrides setting of bad_bits
;    MIN_OVERLAP = Minimum number of pixels in overlap region between two BCDs to
;        include the overlap in the solution. Default for NxN BCD array is N.
;    MAX_FLUX = Maximum image pixel flux to use in overlap analysis. This is intended
;        to allow weeding out bright localized emission from the overlap calculation.
;        There is no default value.
;    FLUX_WEIGHTING : Set to apply a weighting to the pixels equal to the inverse of
;        the flux when calculating the overlap between two BCDs. This is one way to
;        ameliorate the distorting effects of bright emission on the solution.
;    WEIGHT_OVERLAPS : Normally each overlap difference is included in the
;        calculation with equal weight. Set this keyword to weight each overlap
;        according to the area of the overlap (number of pixels). Default not set.
;    WEIGHT_THRESH = Brightness threshold *above* which to apply the overlap area
;        weighting. Usually want uniform weighting for the more quiescent regions.
;        Typically set to a level that flags localized bright emission. Default 
;        is to apply area weighting to all data if WEIGHT_OVERLAPS is set.
;    REDUCE = Array size to perform all calculations, specified as the target array
;        size (assumed an even fraction of the true BCD size), for the purpose of 
;        saving processing time and/or memory.
;        As the overlap precision is also reduced, generally don't use this unless 
;        necessary.
;        Example: REDUCE=128 for 256x256 BCDs will rebin image arrays to 128x128 for
;        the overlap analysis.
;        N. B. The time and memory savings are usually modest.
;    OVERLAPS_PER_BCD : If this is set very accurately as the mean number of overlaps
;        per BCD (or slightly larger), it will prevent the possible need for the 
;        routine to reallocate space for the BCD overlap storage arrays. This may
;        save memory under some circumstances if memory is tight. Default is 60. 
;        N.B. can ignore this keyword in most circumstances.
;    SOURCE_FILE : Text file containing bright point sources, in columnar form:
;            1st col: RA (decimal deg) 
;            2nd col: Dec (decimal deg)
;            3rd col (if present): Flux (any consistent units)
;        Setting this keyword triggers a mode in which the BCDs that contain any
;        of these point sources will be split into multiple segments, the horizontal
;        regions between the sources or between the sources and top or bottom edge.
;        An overlap correction will be calculated for each segment separately.
;        
;
; OFFSET SOLUTION KEYWORDS:
;
;    OUTFILES = vector of filenames to write overlap corrected images to.
;    OUT_DIR = Set to output directory for corrected BCDs. The BCD filenames
;        will be kept the same. The routine will not overwrite so you'll
;        need to explicitly remove existing corrected files in the directory.
;        OUT_DIR must be an existing directory. 
;        OUTFILES takes precedence over OUT_DIR.
;    N_ITERATIONS = Number of solution iterations to perform. Default 1. For
;        increasing values (with a fixed value of STIFFNESS), the overall 
;        solution approaches the "pure overlap" solution. Larger values 
;        usually gives fewer residual level offsets but can introduce an 
;        overall "ramping" in the solution. 
;    STIFFNESS = Parameter specifying the damping factor for n_iterations=1
;        solution. Default is .05, giving a fairly damped solution which preserves
;        overall background levels. For less-damped solution, increase
;        n_iterations rather than decrease stiffness as doing so is much more 
;        robust. Below stiffness=.01 or so the matrix solver can start to behave
;        erratically as the matrix approaches singularity. 
;    WEIGHTS = Set of flags (byte array) for each BCD, set to 0B or 1B. 1B means
;        include BCD as usual in overall solution, 0B means let this BCD float to
;        match the solution of its neighbors. Generally used to flag known
;        outlier BCDs, but setting OUTLIER_THRESH will detect outliers automatically.
;        Use AOR_DEWEIGHT to deweight entire AORs.
;    OUTLIER_THRESH = Deviation threshold for automatic detection and "deweighting" of 
;        outlier BCDs. For each BCD, the median of the absolute overlap differences 
;        with its neighbors is calculated. For BCDs above the threshold (in MJy/sr), 
;        the BCD is deweighted (given weight=0B). The default is not to do this
;        analysis.
;    AOR_DEWEIGHT = List (vector) of AOR code numbers (specified as long ints) 
;        whose BCDs are all to be deweighted (weight=0B). 
;
;
; GROUP OFFSET KEYWORDS:
;
;    GROUP_TAGS: Setting this keyword (or GROUP_AOR) triggers a procedure, 
;        performed prior to the global correction, in which groups of 
;        BCDs are overlap-compared to other groups to determine
;        preliminary group-wide corrections. This is useful for cases in 
;        which groups of BCDs (e.g., AORs) have a consistent background 
;        deviation from the others. This keyword is set as a list (vector) of 
;        integer group tags, one for each BCD in infiles. The tags can 
;        be any ensemble of positive integers. There is no limit to the
;        number of different groups but the practical limit is a few 
;        thousand.
;    GROUP_AOR: Assigns the AOR number of each BCD as the group tag and 
;        triggers the group overlap procedure. Just set as /GROUP_AOR.
;    GROUP_STIFFNESS: Sets the stiffness for the group overlap correction
;        procedure. Default is .03. Set to negative number for stiffness=0
;        (i.e. undamped "pure" overlap algorithm).    
;    GROUP_ANCHORS: List (vector) of group tag numbers for groups that get
;        full weighting in the group corrections procedure. Default is all
;        groups.
;    GROUP_FLOATS: List (vector) of group tag numbers to allow to "float"
;        (i.e., weight=0). Default is all groups have full weight. This is 
;        useful to allow groups that have distinctly deviant backgound levels
;        to freely match the others. 
;    GROUP_ONEPASS: Forces single iteration of group correction procedure,
;        with specified group deweighting (GROUP_ANCHORS or GROUP_FLOATS
;        or none) and group stiffness. Normally there may be more than one
;        pass to optimize group matching:
;        GROUP_ANCHORS is set: one pass by default.
;        GROUP_FLOATS is set: second pass with no group deweighting for 
;            final normalization of levels.
;        No specified deweighting: "zeroth" pass is performed to identify
;            "outlier" groups. These are deweighted for the first pass,
;            and then second pass is performed with no deweighting. For
;            now, "outlier" groups are the half with the higher-magnitude
;            corrections in the zeroth pass.
;    GROUP_OUTLIER_THRESH: This triggers a procedure in which within-group 
;        BCD outliers are flagged and "deweighted", and do not participate
;        in the group overlap correction. Note that these BCDs remain flagged
;        as weight=0B for the remainder of the processing. See OUTLIER_THRESH 
;        for the definition of "outlier".
;    INTRA_GROUP: Setting this triggers a procedure prior to the group
;        overlap correction in which the BCDs in each group are 
;        overlap-corrected only with other BCDs in the group, and the 
;        individual BCD corrections are then applied in the subsequent group 
;        overlap correction. This is intended to smooth out the collective
;        overlap differences between groups. Note that the results are
;        not applied to the data and only affect the outcome through the
;        group overlap corrections.
;    GROUP_EQUALIZE: Normally the group overlap differences are weighted 
;        by the number of BCDs in the overlap region between groups. 
;        Setting this keyword makes all group overlaps contribute 
;        equally to the group offset solution.
;    GROUP_DEWEIGHT: List (vector) of group tag numbers to deweight in their
;        entirety for the final BCD-to-BCD overlap correction. This keyword
;        has no effect on the group overlap correction; this is just a way
;        to specify weight=0 for group-wide sets of BCDs (but is applied only
;        after the group correction). Should be used with care as the spurious
;        ramping can be introduced if the group of BCDs is not well-anchored
;        by overlapping weight=1 data.
;
;
; STRATEGIES:
;    
;    There are two basic specifications necessary: the degree of "damping" 
;    in the solution, and how to handle very bright regions that can typically
;    corrupt the overall solution.
;
;    The default is a highly damped solution, which is intended to preserve
;    overall background levels but can leave residual overlap differences. 
;    The pure "undamped" solution optimizes BCD-to-BCD matching but typically
;    introduces a global ramping to the background levels.
;
;    The damping is controlled with the STIFFNESS and N_ITERATIONS keywords.
;    The stiffness has a default value of .05 and should be left alone
;    in most circumstances, unless it is desired to *increase* the damping
;    effect. The damping effect is reduced by increasing the number of
;    iterations. The general approach should be to start with the damped
;    solution, check the results, and then increase the iterations until
;    the worst residual BCD-to-BCD deviations have been reduced to an 
;    acceptable level. As N_ITERATIONS is increased, the solution approaches
;    the "pure" overlap solution.
;
;    There is no default handling of bright regions as the definition
;    of "bright" depends on the data set. 
;
;    Set OUTLIER_THRESH to a few times higher than typical BCD-to-BCD 
;    deviations seen in the data. This parameter isn't for bright emission 
;    per se but to identify BCDs with significantly deviant values.
;
;    Set MAX_FLUX to the highest value that is generally reliable in the
;    data, or somewhat lower to be conservative.
;
;    Set WEIGHT_OVERLAPS keyword, and set WEIGHT_THRESH to a level above
;    quiescent background levels but below localized bright regions.
; 
;    If problems persist with these settings, then try setting 
;    FLUX_WEIGHTING keyword also. Note that this keyword will increase
;    the processing time by about 50%.
;
;
;    The trickiest situation is where there are significant level differences
;    between large groups of BCDs, for instance AORs taken at different
;    epochs so that there may be a zodiacal variation between overlapping
;    AORs. The basic (damped) overlap correction here is efficient in
;    correcting localized deviations but not very effective at correcting
;    large-scale differences such as this. 
;
;    The simplest solution is to increase N_ITERATIONS until sufficient
;    matching is achieved. The downside is that this can introduce the
;    spurious overall ramping characteristic of the "pure" overlap 
;    solution.
;
;    An alternative is provided that may give better results in some 
;    circumstances. This is to do a pre-correction in which entire groups
;    of BCDs are given a single level adjustment to match other groups of 
;    BCDs. The group-corrected BCDs are then processed with the standard
;    overlap correction.
;    
;    Setting the keyword GROUP_TAGS triggers the group correction
;    mechanism. The BCD groups are identified by numbers (long ints).
;    The group numbers are arbitrary and need not be sequential. 
;    GROUP_TAGS is set as an array (vector) of group numbers, 
;    specifying a group number for each BCD in the input list.
;
;    As specifying a group for each BCD is cumbersome for a large
;    number of BCDs, the keyword GROUP_AOR automatically assigns the AOR 
;    number as the group for each BCD -- that is, the groups are the AORs.
;    It is presumed that this will be the usual usage. If GROUP_AOR
;    is set, GROUP_TAGS should not be set. 
; 
;    The group-to-group (AOR-to-AOR) overlap differences are determined
;    by averaging the individual overlap differences among all the BCDs over 
;    the appropriate pairs of group numbers. The group overlaps are weighted
;    by the number of BCD overlaps between the groups.
;    
;    The default group correction procedure has three steps or passes.
;    In the "zeroth" pass, the corrections are calculated, and the "outlier"
;    groups are defined as the 25% of the groups with the largest-magnitude
;    corrections. In the first pass, the corrections are calculated with
;    these outlier groups given 0 weight (they are allowed to float to match
;    the solution for the remaining groups). The first-pass corrections are
;    applied to the groups. In the second pass, the groups are all given
;    weight=1 and a final set of corrections is calculated and applied to
;    the groups. This last step is to assure that there is at least some
;    global matching of levels over all the groups.
;
;    Judicious selection of the "outlier" groups (AORs) is typically necessary
;    to get good results. The outlier groups can be specified by hand with
;    either the GROUP_FLOATS or GROUP_ANCHORS keywords. With GROUP_FLOATS,
;    known outlier groups are specified as a list (vector). With 
;    GROUP_ANCHORS, the "truth" (non-outlier) groups are specified (and the 
;    remainder are considered outliers). In either case, the "zeroth" pass 
;    is skipped as the outliers are specified. With GROUP_ANCHORS, in addition, 
;    the second (final) correction pass is also skipped, as the "anchor" groups
;    are assumed to be the truth and so a global level matching is not needed.
;    If both GROUP_FLOATS and GROUP_ANCHORS keywords are set, groups listed
;    either in both or in neither are regarded as floats. 
;
;    Other group correction options:
;
;    GROUP_STIFFNESS = Set the damping for the group overlap correction.
;        Default is .03. The solution procedure should be robust enough
;        to assign very small values, say down to .001. Set as a negative 
;        number for the "pure" overlap solution.
;
;    GROUP_OUTLIER_THRESH = Set outlier threshold to ignore BCD-to-BCD
;        overlap in group overlap averaging. Note that only within-group
;        overlaps are considered for this purpose, so this specifies
;        outlier BCDs only within groups (AORs). No default.
; 
;    INTRA_GROUP : If this is set, an overlap correction will be performed
;        on the BCDs for each group separately, and applied to the BCDs
;        for the purpose of determining the group-to-group overlap averages. 
;        The purpose is to help ensure a uniform background level across
;        each group so that the group-to-group overlaps will represent
;        consistent differences. The results are not directly applied to 
;        the final overlap solution. A highly-damped solution (stiffness=.1) 
;        is hardwired to avoid ramping. The default is not to do this
;        group-correction.
;
;    GROUP_EQUALIZE : If this is set, all group-to-group overlaps get 
;        equal weighting in the group-overlap solution. The default is
;        to weight by number of BCD overlaps in the group overlaps.
;
;    GROUP_ONEPASS : If set, there will only be a single correction 
;        pass (i.e., skip the zeroth pass to identify outliers, and
;        also the final equal-weight correction pass).
;
;    GROUP_DEWEIGHT = Set as a list (vector) of groups to be entirely
;        deweighted in the overall solution. Note that this does not
;        affect the group overlap solution. AOR_DEWEIGHT can be used
;        to deweight entire AORs without invoking the group overlap
;        mechanism.
;
;    




;    There are three basic situations: 1) The BCDs exhibit only localized
;    level deviations or minor leg-to-leg deviations, with an overall
;    consistent background (e.g. a single AOR); 2) The bulk of the BCDs
;    have consistent backgrounds but a few groups (AORs) have aberrant
;    background levels; and 3) the BCDs come in groups with varying
;    background levels and no consistent background level.
;
;    1) is the simplest case, the basic correction can be used as-is:
;
;    > afrl_bcd_overlap, infiles, mask_files=..., outfiles=...
;
;    The default stiffness (.05) should give reasonable results in most
;    circumstances. The general idea is that the stiffness should be as low
;    as possible without giving spurious trending, but this will depend
;    on the content of the imaged region (strongly varying brightness 
;    levels tend to result in worse trending). Note that setting the
;    stiffness to about .0001 will approximate the "standard" overlap
;    solution, although for large numbers of BCDs may take awhile to
;    converge.
;
;    Also recommend OUTLIER_THRESH be set at some reasonable value (~1 MJy/sr
;    seems to work fairly well for 24 micron data.)
;
;    Cases 2) and 3) are more complicated as the basic algorithm cannot
;    make reasonable widespread level adjustments. For this we need 
;    to invoke the group correction mechanism. In this process, the
;    BCDs are given single corrections over whole groups, prior to the
;    global offset correction, to put all the groups on the "same page".
; 
;    Setting GROUP_TAGS or GROUP_AOR triggers this mechanism. Setting
;    /GROUP_AOR assigns the AOR number as the group. GROUP_TAGS assigns
;    BCDs to arbitrary groups; it is set as a vector of group numbers,
;    1-to-1 with infiles. The group numbers may be any set of positive
;    integers.
;
;    There are three basic "modes" of group corrections: a) you specify
;    the "truth" groups and allow the others to float to match them; 
;    b) the routine picks the truth groups, and c) all the groups are
;    allowed to float for the best match.
;
;    For a), you set either GROUP_ANCHORS or GROUP_FLOATS depending
;    on whether it's easier to specify the "truth" groups or the ones
;    allowed to float. Set these as an array (vector) of group numbers.
;    Set one or the other but not both.
;
;    For b), set neither GROUP_ANCHORS nor GROUP_FLOATS, and leave the
;    GROUP_STIFFNESS at its default value (.05). What happens is: the
;    routine first does a group offset solution at low stiffness (.01),
;    and selects the third of the groups with the largest-magnitude
;    corrections as the "float" groups. Then the group corrections are
;    determined from this set of "truth" and "float" groups.
;
;    For c), do as b) but set GROUP_STIFFNESS=-1. In this case, the
;    group corrections are the standard overlap correction (all freely
;    floating for the best match). 
;
;
;    If GROUP_ANCHORS is not set, then there is also a final set of
;    corrections determined, with no group "deweighting" and at the
;    specified stiffness, the purpose being to bring all groups to 
;    a common level. The number of groups may be few enough that 
;    non-deweighted groups may only overlap "floating" groups and
;    thus get no correction at all. This last step is to ensure
;    that there is at least some overall matching of groups. Set
;    the GROUP_ONEPASS keyword to suppress this last iteration.
;    (GROUP_ANCHORS are assumed to be "truth".)
;
;
;    SO, for case 2) above, a few anomalous groups, we'd specify the 
;    aberrant groups to float:
;
;    > afrl_bcd_overlap, infiles, corrections, /group_aor, $
;        group_floats=[123456,654321,...]
;
;   or let the routine try to find and "float" the aberrant groups
;   (may work if small percentage of distinctly aberrant groups):
;
;   > afrl_bcd_overlap, infiles, corrections, /group_aor
;
;   If there is one (or a very few) truly problem groups, which you would
;   want to allow to float for the whole process (i.e. global BCD-to-BCD
;   matching also), then set GROUP_DEWEIGHT for that AOR also (this
;   sets weight=0 for BCDs in specified groups for the global matching):
;
;    > afrl_bcd_overlap, infiles, corrections, /group_aor, $
;        group_floats=[123456], group_deweight=[123456]
;
;   Should use GROUP_DEWEIGHT only if the AOR is well-anchored by 
;   overlapping "good" data, else the spurious ramping of the standard
;   overlap solution can occur. Note that GROUP_DEWEIGHT has no effect
;   on the actual group overlap procedure. 
;
;
;   For case 3), all varying background levels, we could either let all
;   the groups float,
;
;   > afrl_bcd_overlap, infiles, corrections, /group_aor, $
;        group_stiffness=-1
;
;   or pick one group (AOR) to be matched to:
;
;   > afrl_bcd_overlap, infiles, corrections, /group_aor, $
;       group_anchors=[123456]
;
;
;
;
;
; NOTES:
;
;    The basic overlap correction algorithm defines chi-squared as the 
;    sum of the squares of the overlap differences with a scalar offset for 
;    each BCD included. Minimization with respect to each offset results 
;    in a condition that the sum of the overlaps differences for each BCD, 
;    including offsets, is 0. In the matrix that represents this condition, 
;    each row represents the condition for 1 BCD; the diagonal element 
;    is the number of overlaps and the off-diagonal elements are -1 for 
;    each overlapping BCD. The right hand side vector element, for that 
;    row, is the sum of the basic overlap differences.
;
;    This matrix is singular; the sum of all the rows but one is equal
;    to the last one, to a scaling factor. Another result is that the
;    solutions typically have some large overall gradient as the solution
;    is consistent with any local gradient and there is no constraint
;    against large-scale excursions of the corrections (only local
;    differences are fixed). 
;
;    This version has a modification that in essence includes the level 
;    adjustments themselves as part of chi-squared. This is intended to 
;    eliminate the large-scale gradients that typically result when applying
;    the basic matching algorithm, and also makes the matrix non-singular.
;    The operational effect is to scale the values of the diagonal elements
;    in the matrix upward by a small fraction, depending on the STIFFNESS
;    parameter, in the range of a few percent. The penalty is that the
;    corrections will be reduced in magnitude by about the same percentage; 
;    if a "proper" correction for a BCD is, say, +.5 MJy/sr, then the 
;    correction would end up maybe +.48 MJy/sr. This should generally be
;    acceptable for the level of corrections necessary, except for large
;    "outliers". For such BCDs, the approach is to "deweight" them: set
;    their WEIGHTS flag equal to zero; they get a correction but it is
;    allowed to "float" to match the neighbor BCDs. Also, can set 
;    OUTLIER_THRESH for automatic detection and deweighting of outliers. 
;    
;    The non-singularity of the matrix allows efficient sparse matrix
;    solver to be used. The built in routine in IDL is called linbcg
;    and is from Numerical Recipes in C. The implementation seems to have 
;    a shortcoming in that it quits without doing anything at some maximum
;    number of total non-zero elements (i.e., individual overlaps), 
;    about 2.5 million, which corresponds to ~45,000 BCDs for the Mipsgal 
;    data. The solution for now is to use other implementations of the
;    linbcg algorithm that have no arbitrary size limit.  Set /INTERNAL 
;    keyword to use version coded in IDL; this is about 10 times slower 
;    but still takes only about 10% of the total processing time. Set 
;    /EXTERNAL to use a stand-alone program called linbcg.x which is assumed 
;    to be in the current directory (or set EXTERNAL= full path to the 
;    executable). This is about as fast as the built-in function.
;
;    Memory requirements:
;
;        The memory used by this routine depends in a complicated way
;        on the number and image size of the BCDs, and the size and shape
;        of the overall subtended region.
;   
;        The memory usage is also influenced by which matrix solver is
;        used; /EXTERN uses about 2/3 the memory of the default built-in
;        solver to store the sparse matrix, and /INTERN uses 1/3 the 
;        memory. But whether this produces a significant difference 
;        overall also depends on the size and shape of the region.
;        
;        Here's a chart for very approximate memory usage in megabytes
;        for 24micron 128x128 images for various region sizes and
;        BCD numbers, for minimum memory usage (/INTERN set). The numbers
;        in parentheses are scaling factors for the worst-case memory
;        usage (built-in matrix solver), although note that the built-in
;        solver likely will not handle more than 50,000 BCDs. /EXTERN
;        will be about midway between the two values.
;
;     deg   N= 2000  10000     25000     50000    100000  200000     500000
;     ---------------------------------------------------------------------
;    .5x.5       51   330(2) 1200(2)   3700(2.5) 
;     1x1        26   145     440(1.5) 1200(2)   3700(2)
;     2x2        15    70     190(1.5)  450(1.5) 1200(2)   3700
;     4x4        10    40      92(1.2)  200(1.3)  460(1.5) 1200
;     2x1        14    75     230(1.5)  600(2)   1800(2.5) 6100
;     4x1         8    38     115(1.6)  310(2)    920(2.2) 3100
;     4x2         8    36     100(1.3)  230(1.5)  610(2)   1900(2.5)
;     10x1        4    16      50(1.5)  125(2)    370(2.5) 1300(2.5)
;     20x1        2     9      25(1.6)   64(2)    190(2)    630(2.5) 3500(3)
;     50x1                               26(2)     80(2)    260(2.5) 1400(3)
;     50x2                                         52(2)    160(2)    760(3)     
;     10x10                                       160(3)    360(1.5) 1300(2)
;
;        For different image size (or using REDUCE option), the
;        numbers should very roughly scale as the number of pixels.
; 
;   Speed:
;
;        With parallel image alignment, the processing time is about 5 minutes per
;        1000 images (on our AFRL Sun machines), and should not depend strongly 
;        on image size. (The processing is arranged so that the time should 
;        increase approximately linearly with the number of BCDs.) 
;
;        With arbitrary alignment, and setting the /EXACT_OVERLAP 
;        keyword, the processing time is ~30 minutes per 1000 BCDs (128x128 images)
;        and should scale approximately with the image size in total pixels. With
;        the /reduce keyword set, the images are binned down to 32x32, and the
;        processing time is ~6 minutes per 1000 images, albeit with some presumed 
;        loss of accuracy in the overlap difference calculation. (Can also set 
;        reduce=size, where size is a divisor of image size, e.g. reduce=64 
;        for 2x reduction of 128x128 images).
;
;        If STIFFNESS is set too low (< .01), then the matrix becomes near-singular,
;        and so standard equation-system solver (SVD operations) is used instead of 
;        the sparse-matrix routines. For 1000 images it takes ~15 minutes to generate
;        a solution, and may increase as fast as N^2. Also, in this case, there is
;        a limit of 10,000 BCDs set as the SVD routines require the full NxN array
;        representation.
;
;    Limitations:
;
;        If the sparse matrix routines are used (stiffness > 1.e-6), then the 
;        limitation in number of BCDs is determined by the limits in the linbcg
;        sparse matrix solver. This procedure seems to quit without attempting 
;        a solution at about 2.5e6 nonzero matrix elements (i.e. total BCD overlaps),
;        which corresponds to about 45,000 BCDs for the Mipsgal data. With the
;        EXTERNAL or INTERNAL versions of linbcg applied, there is really
;        no practical limit on the number of BCDs that can be processed 
;        by the present routine; we could probably handle up to 500,000 BCDs 
;        before memory use becomes taxing (less for regions of high coverage), 
;        and would take a doable 2 days or so to complete. (There is a hardwired 
;        limit of 1e6 BCDs.) It is not know at this time, however, if the 
;        linbcg routine will give a stable solution with this number of BCDs.
;
;        If the SVD solver is used instead (stiffness < 1.e-6), then the 
;        limitation is the size of the matrix since the matrix must be in 
;        full NxN representation. There is a hardwired limit of 10,000 BCDs 
;        (400Mb memory in single precision) in this case, and might take 
;        ~24 hours to process.
;
;        Also, some extreme observing parameters might cause some problems; if there
;        are a very large number of BCDs taken in a small region, then the number
;        of overlaps per BCD will get very large, and this would become problematic 
;        for the routine at some point. 
; 
; 
; PROCEDURES USED:
; 
; FUNCTIONS USED:
;
; HISTORY:
;    31 MAR 2006 Removed multiple iterations and PRELIM keywords, replaced with
;        GROUP offset solution keywords
;    23 MAR 2006 Added multiple iterations, PRELIM_RANGE keyword to deweight based
;        on magnitude of initial corrections.
;     4 MAR 2006 Converted to formal sparse matrix representation throughout.
;    28 Feb 2006 DRM added "external" and "internal" linbcg implementations
;        to allow solutions for very large numbers of BCDs
;        Added save_file and restore_file options
;    25 Feb 2006 DRM Converted to initial difference table representation
;        (allow for outlier detection) and postponed conversion to LHS matrix
;        and RHS vector
;    24 Feb 2006 DRM added sparse matrix output
;    Jan-Feb 2006 DRM Added:
;        "stiffness" parameter -- removes correction gradient and makes matrix nonsingular 
;        Sparse matrix representation and solution -- much faster than SVD
;        Image buffer -- eliminates multiple reads for each image.
;        Processing in sorted order along optimized axis (use with image buffer)
;        Partial vectorization of inner image loop -- big time savings
;        Fast overlap analysis for rectilinearly aligned image sets
;        Image size reduction for faster exact overlap analysis
;    24 May 2005 SJC added mask usage
;    28 May 2005 SJC removed use of pmasks
;    09 June 2005 SJC corrected bug if masks are not being used
;    12 August 2005 SJC changed procedure name and added some comments
;    17 November 2005 SJC made routine image size independent
;-


s = !version
s1 = s.release
s1f = float(s1)
if (s1f lt 5.4) then begin
    print
    print, 'IDL 5.4 or higher necessary.'
    print
    return
endif

if (keyword_set(restore_file)) then begin
    print, 'Restoring environment from file: ' + restore_file
    restore, filename=restore_file
    goto, skip_to_restore
endif

	if (N_params() lt 1 or keyword_set(HELP)) then begin
		return
	endif

        print
        print, 'AFRL BCD OVERLAP   17 November 2008'
        print

	mess = 'AFRL_BCD_OVERLAP:' 
	n = n_elements(infiles)
	if (n gt 1000000L) then begin
            print, 'Maximum BCD count of 10^6 exceeded, aborting.'
            print
            return
	endif

        ; if (keyword_set(REDUCE) and keyword_set(SOURCE_FILE)) then begin
        ;     print, 'Warning: Ignoring REDUCE with SOURCE_FILE set.'
        ;     reduce = 0
        ; endif
        ; should be OK now.

        if (not keyword_set(MAX_FLUX)) then begin
            max_flux = 1.e10
        endif

        if (keyword_set(WEIGHT_THRESH)) then weight_overlaps = 1 
        if (keyword_set(WEIGHT_OVERLAPS) and not keyword_set(WEIGHT_THRESH)) then weight_thresh = -1.e10

        if (not keyword_set(WEIGHTS)) then weights = bytarr(n) + 1B

        ; Hard-wire some previous keywords
        MATCH_ORIENTATION = 1 ; flip BCD arrays/headers to consistent orientation
        EXACT_OVERLAP = 0
        INEXACT_OVERLAP = 0
        EXTERNAL = 0
        INTERNAL = 1 ; use IDL-coded matrix solver
        MAKE_OFFSETS_FILE = 0
        OUTTEXT = 0
        MIN_PIX_FOR_FAST_OVERLAP = 256
        MIN_DIMEN_FOR_FAST_OVERLAP = 8 ; overlap region must be at least 8 pix wide
                                       ; in each dimension


        ; *************************************************************************
        ; Make bright-source RA and Dec lists if indicated from source_file
        ; *************************************************************************

        if (keyword_set(source_file)) then begin
            fstruc = read_ascii(source_file)
            fsz = size(fstruc.field1)
            if (fsz[0] eq 1) then begin
                ; 1-d field array means one entry in source file.
                ra_sources = [fstruc.field1[0]]
                dec_sources = [fstruc.field1[1]]
                if (fsz[1] ge 3) then flux_sources = [fstruc.field1[2]] else flux_sources = 0
                if (fsz[1] ge 6) then code_sources = [fstruc.field1[5]] else code_sources = 0
            end else begin
                ra_sources = reform(fstruc.field1[0,*])
                dec_sources = reform(fstruc.field1[1,*])
                if (fsz[1] ge 3) then flux_sources = [fstruc.field1[2,*]] else flux_sources = 0
                if (fsz[1] ge 6) then code_sources = [fstruc.field1[5,*]] else code_sources = 0
            endelse
            nsrc = n_elements(ra_sources)
            print, nsrc, ' point sources in file.'
        end else begin
            ra_sources = 0
            dec_sources = 0
            flux_sources = 0
            code_sources = 0
        endelse
         

        ; *************************************************************************
        ; Make master lists of BCD ra and dec and other properties.
        ; If source_list is set, the master lists will include a separate
        ; entry for each bright-source array partition.
        ; *************************************************************************

        entries = intarr(n) + 1 ; init to 1 partition entry per BCD.
        if (not keyword_set(source_file)) then begin
            master_ra = fltarr(n)
            master_dec = fltarr(n)
            master_aor = lonarr(n)
            master_rowlo = intarr(n) 
            master_rowhi = intarr(n) 
            master_sourcecol = intarr(n) 
            master_weights = bytarr(n)
            master_rot = fltarr(n)
        end else begin
            ; allow for up to 20 percent duplicate entries
            ; due to BCD partitioning
            ; n1 = round(1.2*float(n))
            ; change 25JUL2007 DRM
            if (n lt 1000) then n1 = 2*n else n1 = max([2000,round(1.2*float(n))])
            master_ra = fltarr(n1)
            master_dec = fltarr(n1)
            master_aor = lonarr(n1)
            infiles_use = strarr(n1)
            if (keyword_set(mask_files)) then mask_files_use = strarr(n1)
            master_rowlo = intarr(n1)
            master_rowhi = intarr(n1) 
            master_sourcecol = intarr(n1)
            master_weights = bytarr(n1)
            master_rot = fltarr(n1)
        endelse

	h0 = headfits(infiles[0])

        ; Get channel number
	ch = sxpar(h0, 'CHNLNUM')

        ; Extract astrometry for all images - procedure does not used the refined
        ; pointing
	print, mess + !stime + ': Extracting astrometry from images'
        print, strtrim(string(n),2) + ' total BCDs.'
	extast, h0, a0

        ; get image width in degrees
        cdelt = sqrt(a0.cd[0,0]^2 + a0.cd[0,1]^2)
        nx1 = sxpar(h0,'NAXIS1')
        ny1 = sxpar(h0,'NAXIS2')
        image_size = float(nx1)*cdelt ; degrees
        dtor = 1.d/360.d*2.d*!dpi

        print, mess + !stime + ': Making global RA and Dec lists.'

        k = 0L  ; running index for partitioned master list

        ythresh = 12 ; lowest y-coordinate to assign a segmenting transition
                    ; (ny-1-ythresh is highest)
        min_ysep = 16

        ; Step through input BCD list.

        for i = 0L,n-1L do begin

               ; print out some completion percentage info
               if (i mod 1000 eq 0 and i gt 0) then print, strtrim(string(i),2)+' BCD headers loaded.'
               p1 = floor(float(i)*10./float(n)) * 10
               p2 = floor(float(i+1)*10./float(n)) * 10
               if (p2 gt p1) then print, '...'+strtrim(string(p2),2)+'%'

            ; Add basic BCD info to lists
            h0 = headfits(infiles[i])
            ra1 = sxpar(h0,'CRVAL1')
            dec1 = sxpar(h0,'CRVAL2')
            aor1 = sxpar(h0,'AORKEY')
            master_ra[k] = ra1 ; N.B. "k" is actual current index if there are partitions
            master_dec[k] = dec1
            master_aor[k] = aor1
            master_weights[k] = weights[i]

            ; calculate rotation angle
            xyad, h0, [float(nx1)/2.,float(nx1)/2.], [0.,float(ny1)], ra_samp, dec_samp
            dra = (ra_samp[1] - ra_samp[0])*cos(dec_samp[0]*dtor)
            ddec = dec_samp[1] - dec_samp[0]
            rotang = atan(-dra,ddec)/dtor ; pos cw from dec "up"
            master_rot[k] = rotang

            ; initialize row range for this BCD
            master_rowlo[k] = 0
            master_rowhi[k] = ny1 - 1
            master_sourcecol[k] = 0

            if (keyword_set(source_file)) then begin

                ; see if any partitioning point sources lie in this BCD.

                infiles_use[k] = infiles[i]
                if (keyword_set(mask_files)) then mask_files_use[k] = mask_files[i]

                ; rough cut in coords first, and consider only actual partitioning sources if applicable
                cosdec = cos(dec1*dtor)
                dra = cosdec*abs(ra_sources - ra1)
                ddec = abs(dec_sources - dec1)
                if (not keyword_set(code_sources)) then $
                    w = where(dra lt image_size and ddec lt image_size,wc)
                if (keyword_set(code_sources)) then $
                    w = where(dra lt image_size and ddec lt image_size and $
                        (code_sources eq 2 or code_sources eq 5),wc)

                if (wc gt 0) then begin

                    ; now see which do actually lie in BCD
                    adxy, h0, ra_sources[w], dec_sources[w], xsources, ysources
                    if (keyword_set(flux_sources)) then fsources = flux_sources[w] 
                    if (keyword_set(code_sources)) then csources = code_sources[w]
                    w = where(xsources ge 0 and xsources le (nx1-1) and $
                              ysources gt ythresh and ysources le (ny1-1-ythresh), wc)

                    if (wc gt 0) then begin
                        ; Extract precise on-BCD list
                        xsources = xsources[w]
                        ysources = ysources[w]
                        if (keyword_set(fsources)) then fsources = fsources[w] 
                        if (keyword_set(csources)) then csources = csources[w] 
                        ; Sort by the source y-axis values
                        ptr = sort(ysources)
                        xsources = xsources[ptr]
                        ysources = ysources[ptr]
                        if (keyword_set(fsources)) then fsources = fsources[ptr] 
                        if (keyword_set(csources)) then csources = csources[ptr] 
                        ; y_last = -1
                        y_last = ythresh - min_ysep ; initialize to allow source at y = ythresh
                        ; step through sources in increasing y-order
                        for j = 0,wc-1 do begin
                            dy = ysources[j] - y_last
                            if (dy gt min_ysep) then begin
                                entries[i] = entries[i] + 1
                                ; Reset upper row limit for current BCD/partition entry to source y.
                                master_rowhi[k] = ysources[j] 
                                ; Now make new entry...
                                k = k + 1L
                                ; Copy basic BCD info
                                master_ra[k] = ra1
                                master_dec[k] = dec1 
                                master_aor[k] = aor1
                                master_weights[k] = weights[i]
                                master_rot[k] = rotang
                                infiles_use[k] = infiles[i] 
                                if (keyword_set(mask_files)) then mask_files_use[k] = mask_files[i]
                                ; Set row limits for this partition
                                ; Initially assume upper limit is top of array, may be reset if more partitions
                                master_rowlo[k] = ysources[j] + 1 
                                master_rowhi[k] = ny1 - 1 
                                master_sourcecol[k] = xsources[j]
                                if (keyword_set(csources)) then if (csources[j] eq 5) then master_sourcecol[k] = -999
                                    ; (flag for "don't do spike filter for this source." normally source code is 2.)
                                y_last = ysources[j]
                            endif 
                        endfor ; loop over on-BCD sources
                    endif ;  any on-BCD sources

                endif ; any near-BCD sources

            endif ; source file exists

            k = k + 1L  ; increment actual array counter

        endfor ; loop over all BCDs


        print, n, ' original BCDs.', format='(i7,a)'
        print, k, ' BCDs for overlaps.', format='(i7,a)'
        ; stop

        if (n_elements(master_ra) gt k) then begin
            ; cut arrays down to true size
            master_ra = master_ra[0:k-1]
            master_dec = master_dec[0:k-1]
            master_aor = master_aor[0:k-1]
            master_weights = master_weights[0:k-1]
            master_rot = master_rot[0:k-1]
            master_rowlo = master_rowlo[0:k-1]
            master_rowhi = master_rowhi[0:k-1]
            master_sourcecol = master_sourcecol[0:k-1]
            infiles_use = infiles_use[0:k-1]
            if (keyword_set(mask_files)) then mask_files_use = mask_files_use[0:k-1]
        endif
        
        ; save original "n" and reset to number of BCD entries, including duplicates
        ; for partitioned BCDs
        n_save = n
        n = k
   
        ; reset BCD and mask lists to total entries, including partition duplicates, 
        ; and save original lists
        if (keyword_set(infiles_use)) then begin 
            infiles_save = temporary(infiles)
            infiles = temporary(infiles_use)
        endif
        if (keyword_set(mask_files_use)) then begin 
            mask_files_save = temporary(mask_files)
            mask_files = temporary(mask_files_use)
        endif


        ; *************************************************************************
        ; Initialize parameters and define local variables
        ; *************************************************************************

        ; see if images are to be reduced for local storage and processing
        ; and set image buffer size from effective image size
        h0 = headfits(infiles[0])
        true_nx = sxpar(h0,'NAXIS1')
        true_ny = sxpar(h0,'NAXIS2')
        reduce_factor = 1
        if (keyword_set(reduce)) then begin
            if (reduce gt 0 and reduce lt 5) then nreduce = 32 else nreduce = reduce
            if (nreduce ge true_nx) then begin
                print, mess + ': Ignoring reduce keyword'
                nreduce = 0
            end else begin
                ; does reduce factor divide into true image size?
                nm = true_nx/nreduce
                if (nreduce*nm ne true_nx) then begin
                    ; nreduce doesn't divide into nx
                    print, mess + ': Improper value of reduce --  exiting'
                    return
                endif
                reduce_factor = nm
            endelse
            if (nreduce gt 0) then begin
                use_nx = nreduce 
                use_ny = nreduce 
            end else begin
                use_nx = true_nx
                use_ny = true_ny
            endelse
        end else begin
            nreduce = 0
            use_nx = true_nx
            use_ny = true_ny
        endelse
        case use_nx of 
            16 : max_buffer = 64000L
            32 : max_buffer = 64000L
            64 : max_buffer = 32000L
            128 : max_buffer = 8000L
            256 : max_buffer = 2000L
            else : max_buffer = round(((128.^2)/float(true_nx)^2)) * 8000L
        endcase

	if (keyword_set(MASK_FILES)) then use_external_masks = 1 else use_external_masks = 0

        ; Mask bits to consider bad for purposes of overlap correction
        ; Use all the bits in an imask
        ; MASK_FILES must be set to utilize these.
 	if (not keyword_set(BAD_BITS) and not keyword_set(BAD_BIT_CODE)) then begin
            bad_bits = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
        endif
        if (keyword_set(BAD_BIT_CODE)) then begin
            bad_bits = [0] ; need some initial array.
            for ib = 0,15 do if ((bad_bit_code and fix(2^ib)) gt 0) then bad_bits = [bad_bits,ib]
            if (n_elements(bad_bits) eq 1) then begin
                print, 'Need some bad bits set.'
                return
            endif
            bad_bits = bad_bits[1:*] ; clip off dummy value at start.
            print, 'bad_bits = ', bad_bits, format='(a,16i3)'
            ; stop
        endif
	
        if (not keyword_set(OVERLAPS_PER_BCD)) then overlaps_per_bcd = 60L

        ; Get image array size, assume same for all images.
        h0 = headfits(infiles[0])
        nx = sxpar(h0, 'NAXIS1') 
        ny = sxpar(h0, 'NAXIS2')
        if (keyword_set(nreduce)) then begin
            ; set for rebinned images
            use_nx = nreduce
            use_ny = nreduce
            nx = use_nx
            ny = use_ny
        endif

        if (not keyword_set(MIN_OVERLAP)) then begin
            ; Set threshold overlap pixel count for inclusion
            pthres = float(nx)
            print, 'Setting min overlap = ', pthres
        end else pthres = min_overlap

	xcorners = [0, 0, nx-1, nx-1]
	ycorners = [0, ny-1, 0, ny-1]


        ; *************************************************************************************
        ; Store table of overlap differences in sparse representation.
        ; This is assembled in the main BCD loop and then later converted to the
        ; LHS matrix (non-zero off-diagonal elements replaced with -1., and the
        ; diagonals +1. for each overlap in matrix row plus stiffness parameter) and 
        ; RHS vector (sums of the difference table rows)
        ;
        ; sparse matrix representation: 2 arrays, sa (float) and ija (long)
        ;   sa[0:n-1] = diagonal values (nxn matrix)
        ;   sa[n] = flag element (value arbitrary)
        ;   sa[n+1:*] = row-by-row nonzero off-diagonal elements, increasing column for each row
        ;   ija[0:n-1] = Starting sa/ija indices for 1st off-diagonal element in row, count from 1
        ;       If no off-diagonal elements, ija[i] = ija[i+1]
        ;   ija[n] = total length of arrays + 1
        ;   ija[n+1:*] = column numbers for off-diagonal elements, count from 1
        ;     Note that ija[0] - 2 = n  (size of matrix)
        ;               ija[ija[0]-1] - 1 = length of sa,ija arrays
        ; *************************************************************************************
        
        estimated_overlaps = overlaps_per_bcd * n ; default is about right for mipsgal
        sa = fltarr(n+1L+estimated_overlaps) ; will be augmented if necessary.
        ija = lonarr(n+1L+estimated_overlaps)
        ; if (keyword_set(weight_overlaps) or keyword_set(source_file)) then wa = bytarr(n+1L+estimated_overlaps) 
        if (keyword_set(weight_overlaps)) then wa = bytarr(n+1L+estimated_overlaps) 
        memory_increment = estimated_overlaps/10  ; augment memory in 10% increments when necessary.

        ; for calculating antisymmetric elements of difference matrix
        ; (explanation below)
        column_markers = lonarr(n)

        ; RHS vector
	vec = fltarr(n)

        ; results for output
        corrections = fltarr(n)
        overlaps = intarr(n)
        max_downstream = lonarr(n) ; for diagnostic purposes, furthest downstream overlapping BCD.

        ; x,y coordinate arrays (for pointing pixels in exact overlap determination)
        ; N.B. nx,ny are use_nx,use_ny.
	xx = indgen(nx) # replicate(1, ny)
	yy = replicate(1, nx) # indgen(ny)
        xxf = float(xx) 
        yyf = float(yy)



        ; *************************************************************************
        ; Sort BCD list along an optimized axis.
        ; When sorted along the optimum axis, overlapping BCDs fall within the
        ; narrowest window. For simple connected regions this is approximately
        ; the axis for which the transverse extent is minimized.
        ;
        ; (First rotate coordinates to be centered at lon=90, lat=0 to avoid pole
        ; effects and any 0/360, -180/+180 discrepancies.)
        ;
        ; Also set image buffer size from the "narrowest overlap window" result.
        ; *************************************************************************

        print, mess + !stime + ': Determining optimum axis for sorting.'

        ; rotate coords so that field center is at (90,0)
        dtor = 1.d/360.d*2.D*!dpi
        ra0 = median(master_ra)
        if (ra0 lt 0.) then ra0 = ra0 + 360.
        dec0 = median(master_dec)
        ; z-rotation to longitude = 90
        q1 = [0.d, 0.d, sin((ra0-90.d)/2.d*dtor), cos((ra0-90.d)/2.d*dtor)]
        ; x-rotation to latitude = 0
        q2 = [sin(dec0/2.*dtor), 0.d, 0.d, cos(dec0/2.*dtor)]
        apply_q, master_ra, master_dec, q1, lon1, lat1
        apply_q, lon1, lat1, q2, lon2, lat2
        
        ; try field rotation intervals in range 0.-90. to find best axis 
        ; orientation for sorting
        min_maxh = 1.e10
        for angle = 0.d,90.d,2.d do begin
            ; y-axis rotation about field center
            q3 = [0.d, sin(angle/2.d*dtor), 0.d, cos(angle/2.d*dtor)]
            apply_q, lon2, lat2, q3, lon3, lat3
            lon3_sorted = lon3(sort(lon3))
            ; Find max possible "downstream" separation between overlapping BCDs
            ; by taking max of histogram of longitude coordinate, with the binsize
            ; being the maximum coordinate separation between overlapping BCDs. Try the
            ; histogram binning with different x-offsets to find "true" maximum.
            maxh = 0L
            for i = 0.,29. do begin 
                ; bin with different lon-offsets, 1/30 BCD width as increment.
                off = i*image_size*1.5/30.
                h = histogram(lon3_sorted+off,binsize=image_size*1.5)
                if (max(h) gt maxh) then maxh = max(h)
            endfor
            ; maxh is approx max separation in sorted list for overlapping BCDs
            ; for this coordinate rotation. See if it is the current global minimum.
            if (maxh lt min_maxh) then begin
                min_maxh = maxh
                master_lon = lon3
                master_lat = lat3
            endif
        endfor

        ; set image buffer size
        ; min_maxh should be close to true max separation,
        ; but add a little padding
        buf_size = min_maxh + min_maxh/10
        if (buf_size gt n) then buf_size = n
        if (buf_size lt 3) then buf_size = 3
        if (buf_size gt max_buffer) then begin
            print, 'Need buffer size greater than maximum allowed, aborting.'
            return
        endif 
        print, mess + !stime + ': Setting image buffer size = '+strtrim(string(buf_size),2)

        master_coslat = cos(master_lat*dtor)

        master_ptr = sort(master_lon)
        sorted_index = master_lon(master_ptr)


        ; *************************************************************************
        ; Define buffer variables
        ; *************************************************************************

        cube = fltarr(nx,ny,buf_size)
        masks = bytarr(nx,ny,buf_size) ; 1B for use, 0B for don't use.
        ra_list = fltarr(buf_size)
        dec_list = fltarr(buf_size)
        lon_list = fltarr(buf_size)
        lat_list = fltarr(buf_size)
        coslat_list = fltarr(buf_size)
        radius_degrees = fltarr(buf_size)
        rotang_list = fltarr(buf_size)
        bcd_nx = intarr(buf_size)
        bcd_ny = intarr(buf_size)
        segmented = bytarr(buf_size) ; flag if a segmented BCD
        buf_index = lonarr(buf_size)
        sort_index = lonarr(buf_size)
	ast = replicate(a0, buf_size) ; a0 returned from extast above.

        n_flip = 0L

        ; *************************************************************************
        ; Fill the image buffer with the first group of BCDs in the sorted list
        ; *************************************************************************

        ; initialize the image buffer
        ; load in 1st buf_size images in sorted list
        print, mess + !stime + ': Initializing image buffer.'
	for i = 0L, buf_size-1L do begin
                    ; print out percentage completion.
                    p1 = floor(float(i)*10./float(buf_size)) * 10
                    p2 = floor(float(i+1)*10./float(buf_size)) * 10
                    if (p2 gt p1) then print, '...'+strtrim(string(p2),2)+'%'
                buf_index[i] = master_ptr[i]  ; ptr to true list index
                sort_index[i] = i             ; ptr to sorted index
                im = readfits(infiles[master_ptr[i]],h0,/SILENT)
                if (not keyword_set(use_external_masks)) then return_valid_pixels, im, mask0
                if (keyword_set(nreduce)) then begin 
                    hrebin, im, h0, rim, rh0, use_nx, use_ny
                    ; N.B. NaNs force rebinned pixel to be NaN.
                    im = rim
                    h0 = rh0
                endif
                nx = sxpar(h0,'NAXIS1')
                ny = sxpar(h0,'NAXIS2')
                rotang = master_rot[master_ptr[i]]
                rowlo = master_rowlo[master_ptr[i]] / reduce_factor
                rowhi = master_rowhi[master_ptr[i]] / reduce_factor
                border = 4 / reduce_factor
                if (keyword_set(match_orientation) and rotang gt -90. and rotang lt 90.) then $
                    flip_bcd = 1 else flip_bcd = 0
                if (flip_bcd) then begin
                    ; N.B. not exact with correction terms but should be OK here.
                    ; also assumes CRPIX is at exact center of array.
                    im = rotate(im,2)
                    sxaddpar, h0, 'CD1_1', -sxpar(h0,'CD1_1')
                    sxaddpar, h0, 'CD2_1', -sxpar(h0,'CD2_1')
                    sxaddpar, h0, 'CD1_2', -sxpar(h0,'CD1_2')
                    sxaddpar, h0, 'CD2_2', -sxpar(h0,'CD2_2')
                    ; Reset rotang to now proper value.
                    if (rotang gt 0.) then rotang = rotang - 180. else rotang = rotang + 180.
                    n_flip = n_flip + 1L
                endif
                cube[*,*,i] = im
                if (not keyword_set(use_external_masks)) then begin
                    if (rowlo gt 0 or rowhi lt (ny-1)) then begin
                        ; mask out section(s) of BCD, for multi-section analysis
                        if (rowlo gt 0) then mask0[*,0:rowlo+border] = 0B
                        if (rowhi lt (ny-1)) then mask0[*,rowhi-border:ny-1] = 0B
                    endif
                    if (keyword_set(nreduce)) then begin
                        mask2 = rebin(float(mask0),use_nx,use_ny)
                        mask0 = byte(ceil(mask2)) ;  0 -> 0,  nonzero -> 1
                    endif
                    if (flip_bcd) then mask0 = rotate(mask0,2)
                    masks[*,*,i] = mask0
                end else begin
                    mask_im = readfits(mask_files[master_ptr[i]],hmask,/SILENT)
                    msize = size(mask_im)
                    if (msize[msize[0]+1] ne 2) then mask_im = fix(round(mask_im)) ; convert to short int
                    return_masked_pixels, mask_im, bad_bits, bptr, bcount
                    mask1 = bytarr(nx,ny) + 1B  ; 1B for use, 0B for don't use.
                    if (bcount gt 0) then mask1(bptr) = 0B
                    if (rowlo gt 0 or rowhi lt (ny-1)) then begin
                        ; mask out section of BCD, for multi-section analysis
                        if (rowlo gt 0) then mask1[*,0:rowlo+border] = 0B
                        if (rowhi lt (ny-1)) then mask1[*,rowhi-border:ny-1] = 0B
                    endif
                    if (keyword_set(nreduce)) then begin
                        mask2 = rebin(float(mask1),use_nx,use_ny)
                        mask1 = byte(ceil(mask2)) ;  0 -> 0,  nonzero -> 1
                    endif
                    if (flip_bcd) then mask1 = rotate(mask1,2)
                    masks[*,*,i] = mask1
                endelse
                ra_list[i] = sxpar(h0,'CRVAL1')
                dec_list[i] = sxpar(h0,'CRVAL2')
                lon_list[i] = master_lon[master_ptr[i]]
                lat_list[i] = master_lat[master_ptr[i]]
                coslat_list[i] = master_coslat[master_ptr[i]]
                bcd_nx[i] = nx
                bcd_ny[i] = ny
                if (rowlo eq 0 and rowhi eq (ny-1)) then segmented[i] = 0b else segmented[i] = 1b
                ; if (transitions[i] eq 0) then segmented[i] = 0b else segmented[i] = 1b
		extast, h0, a0
		a0.cd[0,0] = sxpar(h0, 'CD1_1')
		a0.cd[0,1] = sxpar(h0, 'CD1_2')
		a0.cd[1,0] = sxpar(h0, 'CD2_1')
		a0.cd[1,1] = sxpar(h0, 'CD2_2')
		; ast[i] = a0
		ast[i].cd = a0.cd
		ast[i].cdelt = a0.cdelt
		ast[i].crpix = a0.crpix
		ast[i].crval = a0.crval
		ast[i].ctype = a0.ctype
		ast[i].longpole = a0.longpole
		; ast[i].projp1 = a0.projp1
		; ast[i].projp2 = a0.projp2
		ast[i].pv2[0] = a0.pv2[0]
		ast[i].pv2[1] = a0.pv2[1]
                cdeltx = sqrt(a0.cd[0,0]^2 + a0.cd[0,1]^2) ; degrees per pixel
                radius_degrees[i] = sqrt((float(nx)/2.)^2 + (float(ny)/2.)^2) * cdeltx
                rotang_list[i] = rotang
                ; rotang[i] = atan(a0.cd[0,1]/cdeltx,a0.cd[0,0]/cdeltx)/dtor
	endfor

        ; Initialize buffer pointers
        current_ptr = 0  ; "earliest" bcd in buffer
        global_ptr = buf_size  ; i.e., next sorted index to be loaded into buffer.
        index_tags = lindgen(buf_size)


        ; *************************************************************************
        ; Set up for calculating approximate BCD to BCD pixel coord
        ; transformations, for calculating inexact but relatively accurate
        ; overlap for any relative rotation (faster than exact overlap).
        ; The basic method is to calculate linear transformation coefficients
        ; for transforming BCD1 coords x1,y1 to BCD2 x2,y2, i.e.
        ;     x2 = Ax1 + By1 + C
        ;     y2 = Dx1 + Ey1 + F
        ; The coefficients are determined by least-squares fit to a sampling
        ; of array coords transformed exactly.
        ; Here we are setting up the left-hand-side matrix for the least-squares
        ; solution, which depends only on the sampled array coords.
        ; Note that the left hand side is the same for the 3x3 calculations of
        ; A,B,C and D,E,F separately as they depend only on x1,y1 in the same way.
        ; *************************************************************************

        fx = float(nx) ; N.B. nx and ny are use_nx and use_ny.
        fy = float(ny)
        ; Take 3x3 grid of pixels on the array for sampling exact transformation
        x_samples = double([0.,0.,0.,fx/2.,fx/2.,fx/2.,fx,fx,fx])
        y_samples = double([0.,fy/2.,fy,0.,fy/2.,fy,0.,fy/2.,fy])
        A = dblarr(3,3)
        A[0,0] = total(x_samples*x_samples)
        A[0,1] = total(x_samples*y_samples)
        A[0,2] = total(x_samples)
        A[1,0] = A[0,1]
        A[1,1] = total(y_samples*y_samples)
        A[1,2] = total(y_samples)
        A[2,0] = A[0,2]
        A[2,1] = A[1,2]
        A[2,2] = 9.d
        AINV = invert(A)
        ; The right hand side depends on the specific mapping of coordinates 
        ; for a particular pair of BCDs, e.g. for A,B,C calculation we have
        ; B[0] = total(x_transformed*x_samples)
        ; B[1] = total(x_transformed*y_samples)
        ; B[2] = total(x_transformed)




        ; *************************************************************************
        ; 
        ; Overlap analysis: outer loop over all images
        ;
        ; In this step we are assembling a table of all the image overlap 
        ; differences, with no consideration yet of image weighting.
        ;
        ; The table (matrix) is filled row-by-row. We are only comparing the
        ; "current" BCD with downstream BCDs, so only the "upper triangular"
        ; part of the matrix is filled directly. To build the total sparse 
        ; matrix, we need to fill in the "antisymmetric" portion in the "lower
        ; triangular" part (i.e. diff_ij = -diff_ji). At the start of each
        ; new row "i", we first look for column "i" elements in the previously
        ; filled rows. These become the antisymmetric "lower triangular" 
        ; components of the row, and are filled in before doing any actual  
        ; comparisons.
        ;
        ; *************************************************************************

	print, mess + !stime + ': Overlap analysis -- loop over all images'

        n_off = 0L  ; running count of off-diagonal elements.
        top_index = n + 1L ; pointer to next empty element in sparse arrays 
                           ; (true index, count from 0). 
        n_lower = 0L  ; count off-diagonal elements in lower-triangular part
        n_upper = 0L

        n_inexact = 0L
        n_geom = 0L


	for i = 0L, n-1L do begin

                if (i lt 1000L) then if ((i mod 100L) eq 0 and i gt 0) then $
                    print, i, ' BCDs', format='(i6,a)'
                if (i ge 1000L and i lt 5000L) then if ((i mod 500L) eq 0) then $
                    print, '   ', i, ' BCDs', format='(a,i7,a)'
                if (i ge 5000L and i lt 20000L) then if ((i mod 1000L) eq 0) then $
                    print, '   ', i, ' BCDs', format='(a,i7,a)'
                if (i ge 20000L) then if ((i mod 5000L) eq 0) then $
                    print, '   ', i, ' BCDs', format='(a,i7,a)'

                ; set ija index for this row -- point to 1st element after previous rows'
                ; off diagonal elements even if there end up being no off-diagonal elements in row
                ija[i] = top_index + 1L       ; ija indices are count-from-1


                ; see if sparse arrays need to be expanded in size.
                ; make sure test accounts for possible additions before next test.
                if ((top_index+2L*buf_size) gt n_elements(sa)) then begin
                    ; add another increment of array size.
                    n_old = n_elements(sa)
                    n_new = n_old + memory_increment
                    print, mess+!stime+':Augmenting off-diagonal arrays to ', strtrim(string(n_new),2)
                    tmp = fltarr(n_new)
                    for ii = 0L,n_old-1L do tmp[ii] = sa[ii] ; loop to save memory
                    sa = temporary(tmp)
                    tmp = lonarr(n_new)
                    for ii = 0L,n_old-1L do tmp[ii] = ija[ii]
                    ija = temporary(tmp)
                    ; if (keyword_set(weight_overlaps)) then begin
                    if (keyword_set(wa)) then begin
                        tmp = bytarr(n_new)
                        for ii = 0L,n_old-1L do tmp[ii] = wa[ii]
                        wa = temporary(tmp)
                    endif
                endif


                if (i gt 0L) then begin

                    ; Calculate "antisymmetric" components of this row.
                    ; Look for "column i" elements in previous rows.

                    ; previous row index range
                    lo = max([0L,i-buf_size])  ; need look no further than buf_size from the diagonal.
                    hi = i - 1L

                    marker_subset = column_markers[lo:hi] ; pointers to "next off-diagonal element to consider"
                                                          ; for each previous row.
                    ija_subset = ija[lo+1L:hi+1L]         ; starting indices for "next" rows: to see if
                                                          ; marker_subset values point to true off-diagonal
                                                          ; values or if there are no more elements in row.
                    w = where((marker_subset+1L) lt ija_subset and ija[marker_subset] eq (i+1),wc)
                        ; valid off-diagonal element in row AND column number is current row number (count from 1)
                    if (wc gt 0) then begin
                        ; fill in antisymmetric elements
                        sa[top_index:top_index+wc-1L] = -sa[marker_subset[w]] 
                        ija[top_index:top_index+wc-1L] = lo + w + 1L ; row numbers become column numbers, count from 1
                        ; if (keyword_set(weight_overlaps)) then wa[top_index:top_index+wc-1L] = wa[marker_subset[w]] 
                        if (keyword_set(wa)) then wa[top_index:top_index+wc-1L] = wa[marker_subset[w]] 
                        marker_subset[w] = marker_subset[w] + 1L ; increment markers to next off-diagonal element in row
                        column_markers[lo:hi] = marker_subset ; replace in overall marker list
                        top_index = top_index + wc
                        n_off = n_off + wc
                        n_lower = n_lower + wc
                    endif

                endif

                ; set initial value for the column marker for this row -- the 1st off-diagonal
                ; element to the right of the diagonal if there are any. If there end up 
                ; being no off-diagonal elements to the right, the column marker points to the 
                ; initial index of the next row, which is how we flag no more elements to 
                ; consider in a given row.
                ;
                ; column_markers points to the "next element to consider" for each row, for
                ; the antisymmetric component calculation. It is assumed that the off-diagonal 
                ; elements are in order of increasing column for each row. Thus, for any given
                ; current row i, the column_marker for each earlier row points to either 
                ; 1) a "column i" element, 2) a higher column number element, or 3) no valid
                ; element for the row (none remaining). If 1) we include the antisymmetric 
                ; element and increment the pointer. If 2) or 3) we ignore. column_markers
                ; never points to elements with lower column number than i (for earlier rows)
                ; because those pointers would have been incremented during the processing
                ; of some earlier row.
               
                column_markers[i] = top_index ; column_markers are true indices (count from 0)


                ; for last row, need to calculate antisymmetric elements but no need for comparisons.
                if (i eq n-1L) then goto, skip_comparisons


                ; Read in "current" image

		iim = reform(cube[*,*,current_ptr])
		imask = reform(masks[*,*,current_ptr])
                lon1 = lon_list[current_ptr]
                lat1 = lat_list[current_ptr]
                radius1 = radius_degrees[current_ptr]
                rotangi = rotang_list[current_ptr]

                ; make arrays of lon and lat separations with remaining BCDs in buffer
                ; use "optimized" lon/lat coords, avoid pole effects etc.
                dlon = abs(lon1 - lon_list) * coslat_list
                dlat = abs(lat1 - lat_list)
                ang_sep = sqrt(dlon^2 + dlat^2) ; degrees, not exact but should be OK.
                max_sep = radius1 + radius_degrees ; max separation of bcd centers for possible overlap

                ; extract list of possible overlapping BCDs
                ; ignore "current" BCD and any that are turned off.
                ptr_close = where(ang_sep le max_sep and index_tags ne current_ptr and buf_index ge 0, ptrc)
 
                if (ptrc eq 0) then goto, skip_this_bcd ; no possible overlaps, skip to buffer update

                ; N.B. all images currently need to be same size so this is superfluous
                ; but keep for possible future modification
                nxi = bcd_nx[current_ptr]
                nyi = bcd_ny[current_ptr]

                ; Might want to skip this operation
                if (not keyword_set(exact_overlap) and $
                    not keyword_set(inexact_overlap)) then begin
                    ; assume that BCDs are rectilinearly aligned, can get more precise
                    ; overlap determination from image center separation
                    ; (assume ra,dec in headers are that of image center).
                    ad2xy, ra_list[ptr_close], dec_list[ptr_close], ast[current_ptr], ixclose, iyclose
                    dxclose = abs(float(nxi-1)/2. - ixclose)
                    dyclose = abs(float(nyi-1)/2. - iyclose)
                    w = where(dxclose lt nxi and dyclose lt nyi,wc)
                    if (wc gt 0) then begin
                        ; reset close-bcd pointer and bcd count
                        ptr_close = ptr_close[w]
                        ptrc = wc
                    end else goto, skip_this_bcd
                endif

                ; Calculate celestial coordinates of corners of current image
                xcorners = float([0, 0, nxi-1, nxi-1])
                ycorners = float([0, nyi-1, 0, nyi-1])
		xy2ad, xcorners, ycorners, ast[current_ptr], ircorners, idcorners

                i_use = sort_index[current_ptr] ; index in sorted bcd list, this is index in solution matrices
                    ; this should just be "i"
                current_true = buf_index[current_ptr] ; index in true input list

                ; Note that the difference table is filled in sorted BCD order
                ; This is so we can fill in i,j and j,i elements at the same time.
                ; I.e., we step through the BCD list in sorted order, so we need to
                ; fill the matrix in sorted order to do it efficiently.
                ; (Well, at least the coding is more straightforward.)

                if (keyword_set(exact_overlap)) then begin
                    ; calculate celestial coordinates of all pixels in the current image
                    xy2ad, xx, yy, ast[current_ptr], rr, dd
                end else begin
                    ; calculate celestial coordinates of sample pixels for "inexact"
                    ; (but reasonably accurate) overlap calculation
                    xy2ad, x_samples, y_samples, ast[current_ptr], ra_samples, dec_samples
                endelse 

                
                ; *************************************************************************
                ; Inner loop through possible overlap BCDs
                ; *************************************************************************

                ; need to process potential overlaps in order of increasing j (index 
                ; in sorted list) to fill in the sparse matrix in proper order.

                ptr_jsort = sort(sort_index[ptr_close])

		for jj = 0L,ptrc-1L do begin

                    j = ptr_close[ptr_jsort[jj]]  ; buffer index for this nearby bcd
                    j_use = sort_index[j]         ; index in overall sorted bcd list
                                                  ; this is overlap image index in solution matrices.

                    ; diagnostic, record how far "downstream" in sorted list the
                    ; overlapping BCD is.
                    if ((j_use - i_use) gt max_downstream[current_true]) then $
                        max_downstream[current_true] = j_use - i_use

                    ; get overlap image size and rotation angle
                    ; (currently all BCDs need to be same size)
                    nxj = bcd_nx[j]
                    nyj = bcd_ny[j]
                    rotangj = rotang_list[j]

                    ; See if any corner of the ith image overlaps with the jth image
                    ; N.B. This is an exact calculation for any relative array rotation.

                    ad2xy, ircorners, idcorners, ast[j], jxcorners, jycorners
                    ptr = where(jxcorners ge 0. and jxcorners lt float(nxj)-1. and $
                        jycorners ge 0. and jycorners lt float(nyj)-1., jccount)

                    ; ... or, if any corner of the jth image overlaps with the ith image
                    ; Calculate celestial coordinates of corners of jth image

                    xy2ad, xcorners, ycorners, ast[j], jrcorners, jdcorners
                    ad2xy, jrcorners, jdcorners, ast[current_ptr], ixcorners, iycorners
                    ; Apply 1 pixel "padding"
                    ptr = where(ixcorners ge -1. and ixcorners lt float(nxi) and $
                        iycorners ge -1. and iycorners lt float(nyi), iccount)

                    ; *********************************************************************
                    ; If either image has a corner in the other image then there is overlap
                    ; and we need to calculate the overlap value. 
                    ; Hmmm, this criterion might not be exhaustive...
                    ; *********************************************************************

                    if (jccount gt 0 or iccount gt 0) then begin

                        ; Read in overlap image (jth image in buffer)

                        n_geom = n_geom + 1L

                        jim = reform(cube[*,*,j])
	                jmask = reform(masks[*,*,j])

                        ; ***********************************************
                        ; See if "inexact" overlap needs to be done.
                        ; This check is for a rotation between the BCDs
                        ; ***********************************************

                        inexact_for_this = 0

                        if (not keyword_set(exact_overlap)) then begin
                            ; Calculate coefficients for linear transformation
                            ; for "inexact" overlap, and check the rotation components.
                            ad2xy, ra_samples, dec_samples, ast[j], x_mapped, y_mapped
                            B1 = dblarr(3)
                            B2 = dblarr(3)
                            B1[0] = total(x_mapped*x_samples)
                            B1[1] = total(x_mapped*y_samples)
                            B1[2] = total(x_mapped)
                            B2[0] = total(y_mapped*x_samples)
                            B2[1] = total(y_mapped*y_samples)
                            B2[2] = total(y_mapped)
                            C1 = AINV # B1
                            C2 = AINV # B2
                            if (not keyword_set(inexact_overlap)) then begin
                                ; Look at rotation components to see if there's skew.
                                c1amp = sqrt(C1[0]^2 + C1[1]^2)
                                c2amp = sqrt(C2[0]^2 + C2[1]^2)
                                maxc1 = max(abs([C1[0],C1[1]]))
                                maxc2 = max(abs([C2[0],C2[1]]))
                                rmax = max([maxc1/c1amp,maxc2/c2amp])
                                if (rmax lt 0.999) then begin
                                    ; relative rotation is > 2 degrees or so.
                                    inexact_for_this = 1
                                endif
                            endif
                        endif

                        ; ****************************************************************
                        ; Calculate the overlap regions.
                        ; ****************************************************************

                        if (not keyword_set(exact_overlap) and $
                            not keyword_set(inexact_overlap) and $
                            not keyword_set(inexact_for_this)) then begin

                            ; Fast overlap calculation.
                            ; Assume BCDs are "parallel" and overlap region is thus rectilinear
                            ; Extract overlap regions and calculate difference

                            ; calculate overlap region in overlap image
                            ; Expand boundaries by 1 pixel (23FEB2007 DRM)
                            jxmin = round(min([nxj-1,max([0,min(jxcorners-1.)])]))
                            jxmax = round(max([0,min([nxj-1,max(jxcorners+1.)])]))
                            jymin = round(min([nyj-1,max([0,min(jycorners-1.)])]))
                            jymax = round(max([0,min([nyj-1,max(jycorners+1.)])]))

                            ; calculate overlap region in "current" image
                            ; Expand boundaries by 1 pixel (23FEB2007 DRM)
                            ixmin = round(min([nxi-1,max([0,min(ixcorners-1.)])]))
                            ixmax = round(max([0,min([nxi-1,max(ixcorners+1.)])]))
                            iymin = round(min([nyi-1,max([0,min(iycorners-1.)])]))
                            iymax = round(max([0,min([nyi-1,max(iycorners+1.)])]))

                            ; see if overlap region is large enough to do fast overlap
                            dcol = ixmax - ixmin
                            drow = iymax - iymin
                            min_dimen = min([dcol,drow])
                            if (min_dimen lt MIN_DIMEN_FOR_FAST_OVERLAP) then begin
                                inexact_for_this = 1
                                goto, skip_fast_overlap
                            endif

                            ; Extract overlap regions.
                            iim_over = iim[ixmin:ixmax,iymin:iymax]
                            imask_over = imask[ixmin:ixmax,iymin:iymax]
                            jim_over = jim[jxmin:jxmax,jymin:jymax]
                            jmask_over = jmask[jxmin:jxmax,jymin:jymax]

                            ; Make sure the overlap difference is calculated only where there's
                            ; valid data in both overlap regions.
                            ; This is necessary if there are large masked regions in the array.
                            ; 30 OCT 2006 DRM

                            ; Method: set overlap regions for both images to identical pixel dimensions
                            ; and then require that corresponding pixels in both have valid
                            ; values.

                            ; check for 1-pixel wide overlaps, and make explicitly 2-D
                            si = size(iim_over)
                            if (si[0] eq 1) then begin
                                ; reform as 2-D array
                                iim_over = reform(iim_over,ixmax-ixmin+1,iymax-iymin+1)
                                imask_over = reform(imask_over,ixmax-ixmin+1,iymax-iymin+1)
                                si = size(iim_over)
                            endif
                            sj = size(jim_over)
                            if (sj[0] eq 1) then begin
                                ; reform as 2-D array
                                jim_over = reform(jim_over,jxmax-jxmin+1,jymax-jymin+1)
                                jmask_over = reform(jmask_over,jxmax-jxmin+1,jymax-jymin+1)
                                sj = size(jim_over)
                            endif

                            ; extract same-sized overlap regions (smaller of each dimension)
                            ; (might want to determine more exact correspondence...)
                            nxo = min([si[1],sj[1]])
                            nyo = min([si[2],sj[2]])
                            iim_over = iim_over[0:nxo-1,0:nyo-1]
                            jim_over = jim_over[0:nxo-1,0:nyo-1]
                            imask_over = imask_over[0:nxo-1,0:nyo-1]
                            jmask_over = jmask_over[0:nxo-1,0:nyo-1]

                            ; use only good data pixels in both overlap regions
                            wij = where(imask_over eq 1B and jmask_over eq 1B and $
                                        iim_over le max_flux and jim_over le max_flux,wijc)
                            if (wijc gt pthres) then begin
                                iim_overlap = iim_over[wij]
                                jim_overlap = jim_over[wij]
                                gcount = wijc
                            end else gcount = 0

                            if (wijc lt MIN_PIX_FOR_FAST_OVERLAP) then inexact_for_this = 1
                            
                            skip_fast_overlap:

                        endif 

                        if (keyword_set(exact_overlap) or keyword_set(inexact_overlap) or $
                            keyword_set(inexact_for_this)) then begin

                            ; Calculate more precise overlap region, to handle relative rotations,
                            ; slower than fast overlap.

                            if (keyword_set(inexact_overlap) or keyword_set(inexact_for_this)) then begin
                                ; apply the linear transformation to all the array pixels
                                jxx = C1[0]*xxf + C1[1]*yyf + C1[2]
                                jyy = C2[0]*xxf + C2[1]*yyf + C2[2]
                                n_inexact = n_inexact + 1L
                            endif

                            if (keyword_set(exact_overlap)) then begin
                                ; explicitly calculate BCD j pixel coordinates of all current image pixel coords
                                ; much slower than inexact or rectilinear overlap.
                                ad2xy, rr, dd, ast[j], jxx, jyy
                            endif

                            ; round mapped pixel coord to nearest target pixel (15AUG2008 DRM)
                            jxx = round(jxx)
                            jyy = round(jyy)

                            ptr = where(jxx ge 0 and jxx le nx-1 and $
                                jyy ge 0 and jyy le ny-1, count)

                            ; Need to refine the overlap by applying the mask
                            ; Is there a more efficient way to do this?
                            if (count gt 0) then begin
                                ; gptr = where((imask[xx[ptr], yy[ptr]] eq 1B) and $
                                ;     (jmask[jxx[ptr], jyy[ptr]] eq 1B), gcount)
                                ; added max_flux test 28OCT2006 DRM
                                gptr = where((imask[xx[ptr], yy[ptr]] eq 1B) and $
                                         (jmask[jxx[ptr], jyy[ptr]] eq 1B) and $
                                         (iim[xx[ptr], yy[ptr]] le max_flux) and $  
                                         (jim[jxx[ptr], jyy[ptr]] le max_flux), gcount)
                                if (gcount gt pthres) then begin
                                    iim_overlap = iim[xx[ptr[gptr]], yy[ptr[gptr]]]
                                    jim_overlap = jim[jxx[ptr[gptr]], jyy[ptr[gptr]]]
                                endif
                            end else gcount = 0

                        endif

                        ; If number of pixels in overlap is greater than threshold,
                        ; add to difference table. (N.B. no consideration of image
                        ; weights here, just compile all overlap differences.)
                        ; This adds to the "upper triangular" part of the matrix.
                        ; The antisymmetric "lower triangular" part is added at
                        ; the start of processing each row. 

                        if (gcount gt pthres) then begin

                            ; **********************************************
                            ; Calculate overlap difference and weight factor
                            ; **********************************************

                            if (keyword_set(flux_weighting)) then begin
                                ; Take weighted average of pixel differences
                                ; Weight is inverse pixel brightness
                                ; Also weed out outlier pixel differences from average
                                d = iim_overlap - jim_overlap
                                med = median(d)
                                mom = moment(d)
                                rms = sqrt(mom[1])
                                wgood = where(abs(d-med) le 3.*rms and iim_overlap gt 0.)
                                wgt = 1./(iim_overlap[wgood] + 10.) ; damp small image values.
                                overlap_difference = total(d[wgood]*wgt)/total(wgt)
                                overlap_med = median(iim_overlap)
                            end else begin
                                medi = median(iim_overlap)
                                medj = median(jim_overlap)
                                overlap_difference = medi - medj
                                ; overlap_med = (medi + medj)/2. 
                                overlap_med = max([medi,medj])
                            endelse
                            use_weight = float(gcount) / (float(nxi)*float(nyi))
                            if (use_weight gt 1.) then use_weight = 1.

                            ; **********************************************
                            ; Add difference to matrix
                            ; **********************************************

                            sa[top_index] = overlap_difference
                            ija[top_index] = j_use + 1L  ; column number, count from 1
                            if (keyword_set(wa)) then begin
                                if (overlap_med gt weight_thresh) then begin
                                    wa[top_index] = byte(floor(use_weight*127.))
                                end else wa[top_index] = 127b
                            endif

                            top_index = top_index + 1L ; pointer to next unused element in sa,ija arrays
                            n_off = n_off + 1L
                            n_upper = n_upper + 1L

                        endif ; sufficient good-data overlap for inclusion

                    endif ; any geometric overlap at all

		endfor  ; inner loop over nearby bcds

                skip_this_bcd:

                ; ********************************************************************
                ; update the image buffer:
                ; load in next BCD from global sorted list into "current" position,
                ; and increment "current" position to next location in buffer
                ; ********************************************************************

                if (global_ptr lt n) then begin
                    buf_index[current_ptr] = master_ptr[global_ptr] ; true input list index
                    sort_index[current_ptr] = global_ptr            ; sorted list index
                    im = readfits(infiles[master_ptr[global_ptr]],h0,/SILENT)
                    if (not keyword_set(use_external_masks)) then return_valid_pixels, im, mask0
                    if (keyword_set(nreduce)) then begin 
                        hrebin, im, h0, rim, rh0, use_nx, use_ny
                        ; N.B. NaNs force rebinned pixel to be NaN.
                        im = rim
                        h0 = rh0
                    endif
                    nx = sxpar(h0,'NAXIS1')
                    ny = sxpar(h0,'NAXIS2')
                    rotang = master_rot[master_ptr[global_ptr]]
                    if (keyword_set(match_orientation) and rotang gt -90. and rotang lt 90.) then $
                        flip_bcd = 1 else flip_bcd = 0
                    if (flip_bcd) then begin
                        ; N.B. not exact with correction terms but should be OK here.
                        ; Also assumes CRPIX is exact center of array.
                        im = rotate(im,2)
                        sxaddpar, h0, 'CD1_1', -sxpar(h0,'CD1_1')
                        sxaddpar, h0, 'CD2_1', -sxpar(h0,'CD2_1')
                        sxaddpar, h0, 'CD1_2', -sxpar(h0,'CD1_2')
                        sxaddpar, h0, 'CD2_2', -sxpar(h0,'CD2_2')
                        ; Reset rotang to now proper value
                        if (rotang gt 0.) then rotang = rotang - 180. else rotang = rotang + 180.
                        n_flip = n_flip + 1L
                    endif
                    cube[*,*,current_ptr] = im
                    rowlo = master_rowlo[master_ptr[global_ptr]] / reduce_factor
                    rowhi = master_rowhi[master_ptr[global_ptr]] / reduce_factor
                    border = 4 / reduce_factor
                    if (not keyword_set(use_external_masks)) then begin
                        if (rowlo gt 0 or rowhi lt (ny-1)) then begin
                            ; mask out section of BCD, for multi-section analysis
                            if (rowlo gt 0) then mask0[*,0:rowlo+border] = 0B
                            if (rowhi lt (ny-1)) then mask0[*,rowhi-border:ny-1] = 0B
                        endif
                        if (keyword_set(nreduce)) then begin
                            mask2 = rebin(float(mask0),use_nx,use_ny)
                            mask0 = byte(ceil(mask2)) ;  0 -> 0,  nonzero -> 1
                        endif
                        if (flip_bcd) then mask0 = rotate(mask0,2)
                        masks[*,*,current_ptr] = mask0
                    end else begin
                        mask_im = readfits(mask_files[master_ptr[global_ptr]],hmask,/SILENT)
                        msize = size(mask_im)
                        if (msize[msize[0]+1] ne 2) then mask_im = fix(round(mask_im)) ; convert to short int
                        return_masked_pixels, mask_im, bad_bits, bptr, bcount
                        mask1 = bytarr(nx,ny) + 1B  ; 1B for use, 0B for don't use.
                        if (bcount gt 0) then mask1(bptr) = 0B
                        if (rowlo gt 0 or rowhi lt (ny-1)) then begin
                            ; mask out section of BCD, for multi-section analysis
                            if (rowlo gt 0) then mask1[*,0:rowlo+border] = 0B
                            if (rowhi lt (ny-1)) then mask1[*,rowhi-border:ny-1] = 0B
                        endif
                        if (keyword_set(nreduce)) then begin
                            mask2 = rebin(float(mask1),use_nx,use_ny)
                            mask1 = byte(ceil(mask2)) ;  0 -> 0,  nonzero -> 1
                        endif
                        if (flip_bcd) then mask1 = rotate(mask1,2)
                        masks[*,*,current_ptr] = mask1
                    endelse
                    ra_list[current_ptr] = sxpar(h0,'CRVAL1')
                    dec_list[current_ptr] = sxpar(h0,'CRVAL2')
                    lon_list[current_ptr] = master_lon[master_ptr[global_ptr]]
                    lat_list[current_ptr] = master_lat[master_ptr[global_ptr]]
                    coslat_list[current_ptr] = master_coslat[master_ptr[global_ptr]]
                    bcd_nx[current_ptr] = nx
                    bcd_ny[current_ptr] = ny
                    if (rowlo eq 0 and rowhi eq (ny-1)) then $
                        segmented[current_ptr] = 0b else segmented[current_ptr] = 1b
		    extast, h0, a0
		    a0.cd[0,0] = sxpar(h0, 'CD1_1')
		    a0.cd[0,1] = sxpar(h0, 'CD1_2')
		    a0.cd[1,0] = sxpar(h0, 'CD2_1')
		    a0.cd[1,1] = sxpar(h0, 'CD2_2')
		    ast[current_ptr].cd = a0.cd
		    ast[current_ptr].cdelt = a0.cdelt
		    ast[current_ptr].crpix = a0.crpix
		    ast[current_ptr].crval = a0.crval
		    ast[current_ptr].ctype = a0.ctype
		    ast[current_ptr].longpole = a0.longpole
		    ; ast[current_ptr].projp1 = a0.projp1
	            ; ast[current_ptr].projp2 = a0.projp2
		    ast[current_ptr].pv2[0] = a0.pv2[0]
		    ast[current_ptr].pv2[1] = a0.pv2[1]
                    cdeltx = sqrt(a0.cd[0,0]^2 + a0.cd[0,1]^2) ; degrees per pixel
                    radius_degrees[current_ptr] = sqrt((float(nx)/2.)^2 + (float(ny)/2.)^2) * cdeltx
                    rotang_list[current_ptr] = rotang
                    ; rotang[current_ptr] = atan(a0.cd[0,1]/cdeltx,a0.cd[0,0]/cdeltx)/dtor
                    global_ptr = global_ptr + 1L
                end else begin
                    ; we've loaded all the BCDs, so just "turn off" the current BCD
                    ; from further consideration
                    buf_index[current_ptr] = -999L
                    sort_index[current_ptr] = -999L
                endelse

                current_ptr = current_ptr + 1L
                if (current_ptr eq buf_size) then current_ptr = 0L

                skip_comparisons:
                
	endfor  ; outer bcd loop

        print, mess + !stime + ': ' + strtrim(string(n_geom)) + ' geometric overlaps processed.'
        print, mess + !stime + ': ' + strtrim(string(n_inexact)) + ' inexact overlaps calculated.'
        print, mess + !stime + ': ' + strtrim(string(n_flip)) + ' inverted BCDs.'

        ; N.B. The bcd ordering in the matrices is the sorted order, not input order.


        ; *************************************************************************
        ; Set "flag element" in sparse matrix arrays
        ; *************************************************************************

        sa[n] = 0.
        ija[n] = top_index + 1L


        ; *************************************************************************
        ; No overlaps? Just quit.
        ; *************************************************************************

        if (n_off eq 0) then begin
            print, mess + !stime + ': No overlaps found, exiting.'
            goto, skip_solution
        endif

        if (n_off gt 2200000L and $
            not keyword_set(internal) and not keyword_set(external)) then begin
            print, mess + !stime + $
                ': Warning, built-in function linbcg may have problems with this size matrix.'
            print, mess + !stime + ': May need to set /external or /internal for solution.'
        endif


        ; *************************************************************************
        ; Save point for save_file 
        ; Entry point for restore_file
        ; *************************************************************************

        if (keyword_set(save_file)) then begin
            save, filename=save_file
            print, mess + !stime + ': Exiting after saving environment in ' + save_file
            return
        endif

        skip_to_restore:


        ; *************************************************************************
        ; Now process offset-solution input keywords
        ; *************************************************************************

        max_svd_bcds = 10000L ; max number of BCDs for SVD solution (i.e. very low stiffness)
        MAX_TABLE = 2000L ; max size for returned difference table 

        ; parameter for including overlap corrections themselves as part of chi-squared.
        if (not keyword_set(STIFFNESS)) then begin
            print, 'Setting stiffness to .05'
            stiffness = .05 
        endif
        min_stiffness = 1.e-6
        if (stiffness[0] lt min_stiffness) then begin
            if (n gt max_svd_bcds) then begin
                print, 'Stiffness too small, must be greater than ' + $
                    strtrim(string(min_stiffness),2) + ' for sparse matrix operations.'
                print, 'Number of BCDs too large for SVD operations.'
                return
            end else begin
                print, 'Stiffness too small, will use SVD for solution, may be very slow.'
                use_svd = 1
                if (keyword_set(external) or keyword_set(internal)) then $
                    print, 'EXTERNAL and INTERNAL not valid for SVD solution, ignoring.'
                if (keyword_set(external)) then external = 0
                if (keyword_set(internal)) then internal = 0
            endelse
            if (stiffness[0] lt 0.) then stiffness = 0.
        end else use_svd = 0

        ; Set image weight vector
        ; if (not keyword_set(WEIGHTS)) then weights = bytarr(n) + 1B
        ; ngood = n_elements(weights)

        if (not keyword_set(N_ITERATIONS)) then begin
            print, 'Setting single correction iteration.'
            n_iterations = 1
        endif

        if (keyword_set(SVD)) then begin
            if (n le max_svd_bcds) then use_svd = 1
            if (n gt max_svd_bcds) then print, 'Too many BCDs for SVD solution, ignoring.'
        endif
           
        if (keyword_set(GROUP_TAGS) or keyword_set(GROUP_AOR)) then begin
             ; set up for group correction procedure
             if (keyword_set(GROUP_AOR)) then begin
                 group_tags = master_aor
             end else begin
                 if (n ne n_save) then begin
                     ; need to reset group tag list to include duplicates
                     group_tags1 = lonarr(n)
                     k = 0L
                     for i = 0L,n_save-1L do begin
                         for j = 0L,entries[i]-1L do begin
                             group_tags1[k] = group_tags[i]
                             k = k + 1L
                         endfor
                     endfor
                     group_tags_save = group_tags
                     group_tags = group_tags1
                 endif
             endelse
             ; make list of unique tags
             group_tags_sort = group_tags[sort(group_tags)]
             group_taglist = group_tags_sort[uniq(group_tags_sort)]
             ngroups = n_elements(group_taglist)
             ; make map from BCD index to tag index
             ; inelegant method for now.
             group_index = lonarr(n)
             for i = 0L,n-1L do begin
                 w = where(group_taglist eq group_tags[i])
                 group_index[i] = w[0]
             endfor
             ; make group_weights array
             group_weights = bytarr(ngroups) + 1B
             if (keyword_set(GROUP_ANCHORS)) then begin
                 group_weights[*] = 0B
                 n_anch = n_elements(group_anchors)
                 for i = 0L,n_anch-1L do begin
                     w = where(group_taglist eq group_anchors[i],wc)
                     if (wc gt 0) then group_weights[w[0]] = 1B else $
                         print, '   Ignoring tag = ', strtrim(string(group_anchors[i]),2),' not in BCD list.'
                 endfor
             endif
             if (keyword_set(GROUP_FLOATS)) then begin
                 n_float = n_elements(group_floats)
                 for i = 0L,n_float-1L do begin
                     w = where(group_taglist eq group_floats[i],wc)
                     if (wc gt 0) then group_weights[w[0]] = 0B else $
                         print, '   Ignoring tag = ', strtrim(string(group_floats[i]),2),' not in BCD list.'
                 endfor
             endif
             if (not keyword_set(GROUP_STIFFNESS)) then group_stiffness = .01
             if (GROUP_STIFFNESS lt 0.) then group_stiffness = 0.
        endif
             
 
        ; *************************************************************************
        ; Make full-array difference table if asked for
        ; *************************************************************************

        if (arg_present(diagnostics) and n le MAX_TABLE) then begin
            top_index = ija[n] - 1L
            m_sparse = {sa:dblarr(top_index), ija:lon64arr(top_index)}
            for ii = 0L,top_index-1L do begin ; explicit loop to save memory
                m_sparse.sa[ii] = double(sa[ii])
                m_sparse.ija[ii] = long64(ija[ii])
            endfor
            full_table = fulstr(temporary(m_sparse))
        end else full_table = 0
        

        ; *************************************************************************
        ; Large-outlier detection for automatic BCD deweighting
        ; 4-sigma outlier detection for overlap deletion
        ; Criteria:
        ; BCD deweighting: Median of absolute differences is greater than 
        ;     outlier_thresh
        ; Overlap deletion: Absolute of an overlap difference is 4-sigma deviation 
        ;     from median of absolute differences for both BCDs separately 
        ; N.B. overlap deletion requires WEIGHT_OVERLAPS to be set (uses weights
        ; matrix wa to delete overlap)
        ; *************************************************************************

        n_deweighted = 0L ; count deweighted BCDs
        deweighted = bytarr(n) ; 0B init assumed
        n_deleted = 0L ; count deleted overlaps

        median_deviationsx = fltarr(n)
        mean_deviations = fltarr(n)

        if (keyword_set(outlier_thresh) or keyword_set(wa)) then begin
            print, mess + !stime + ': Performing BCD/overlap outlier analysis.'

            ; Step through all BCDs
            for i = 0L,n-1L do begin
                istart = ija[i] - 1L
                istop = ija[i+1] - 2L
                if (istart le istop) then begin

                    ; there are some off-diagonal elements (overlaps) to consider.
                    sa_this = sa[istart:istop]
                    ija_this = ija[istart:istop] - 1L ; columns, counting from 0
                    if (keyword_set(wa)) then wa_this = wa[istart:istop]
                    nsa = n_elements(sa_this)
                    medi = median(abs(sa_this))
                    median_deviationsx[master_ptr[i]] = medi
                    if (keyword_set(outlier_thresh)) then begin
                        ; See if BCD deweighting criterion is met.
                        if (medi ge outlier_thresh) then begin
                            ; weights[master_ptr[i]] = 0B
                            master_weights[master_ptr[i]] = 0B ; fix 21SEP2007 DRM
                            n_deweighted = n_deweighted + 1L
                            deweighted[master_ptr[i]] = 1B
                        endif
                    endif
                    if (nsa ge 2) then begin
                        ; Calculate mean overlap difference
                        momi = moment(sa_this)
                        mean_deviations[master_ptr[i]] = momi[0]
                    end else mean_deviations[master_ptr[i]] = medi

                    ; Now do overlap-outlier analysis - deweight overlaps that are 
                    ; 4-sigma outliers for both BCDs 
                    if (nsa gt 2 and keyword_set(wa) and 0) then begin
                        ; Calculate mean and rms of absolute overlap differences for this BCD
                        momi = moment(abs(sa_this))
                        rmsi = sqrt(momi[1])
                        devi = abs(abs(sa_this) - medi)
                        ; Find 4-sigma outliers
                        ptr_outlier = where(devi gt 4.*rmsi,ptrc)
                        if (ptrc gt 0) then for j = 0,ptrc-1L do begin ; step through outliers
                            adiff = abs(sa_this[ptr_outlier[j]]) ; absolute overlap difference
                            jcol = ija_this[ptr_outlier[j]] ; column index of this outlier, count from 0
                            jstart = ija[jcol] - 1L ; extract overlap indices for overlapping BCD
                            jstop = ija[jcol+1] - 2L
                            if (jstart le jstop) then begin
                                ; Extract all overlap differences and column indices for overlapping BCD
                                sa_that = sa[jstart:jstop]
                                ija_that = ija[jstart:jstop] - 1L ; count from 0
                                nsa_that = n_elements(sa_that)
                                if (nsa_that ge 2) then begin
                                    ; Identify matching overlap entry for "current" BCD
                                    ptr_match = where(ija_that eq i,ptrc) ; should always be a match
                                    if (ptrc eq 1) then begin
                                        ; Calculate median and rms for overlapping BCD's overlaps
                                        medj = median(abs(sa_that))
                                        momj = moment(abs(sa_that))
                                        rmsj = sqrt(momj[1])
                                        ; See if overlap difference is 4-sigma deviation for other BCD also
                                        devj = abs(adiff - medj)
                                        if (devj gt 4.*rmsj) then begin
                                            ; Deweight overlap for both BCDs
                                            wa_this[ptr_outlier[j]] = 0B ; will be replaced in wa array later.
                                            wa[jstart+ptr_match[0]] = 0B
                                            n_deleted = n_deleted + 1L
                                        endif
                                    endif ; else some error...
                                endif
                            endif
                        endfor
                        wa[istart:istop] = wa_this
                    endif
                endif
            endfor

            print, mess + !stime + ': Deweighted BCDs     = ' + strtrim(string(n_deweighted),2)
            if (keyword_set(wa)) then print, mess + !stime + ': Deweighted overlaps = ' + strtrim(string(n_deleted),2)

        endif
         
        ; Copy to output parameter if not already set as difference table
        if (not keyword_set(median_deviations) and keyword_set(median_deviationsx)) then $
            median_deviations = median_deviationsx else median_deviations = 0



        ; *************************************************************************
        ; Do group overlap correction if specified.
        ; (i.e., determine scalar corrections for each group from group-to-group
        ; BCD overlaps)
        ; 
        ;   1. Do within-group outlier detection and BCD deweighting (optional)
        ;   2. Do within-group BCD overlap matching for preliminary correction (optional) 
        ;   3. Assemble group overlap difference table from BCD difference table.
        ;   4. Loop through passes of group offset solutions:
        ;      Pass 0: Do initial solution with no group deweighting to find 
        ;              group outliers. Deweight the outlier groups for Pass 1.
        ;              (Skip if GROUP_ANCHORS or GROUP_FLOATS is set)
        ;      Pass 1: Do solution with specified group deweighting and apply 
        ;              offsets to group difference table.
        ;      Pass 2: Do solution with updated group difference table and no
        ;              group deweighting for final matching of levels.
        ;              (Skip if GROUP_ANCHORS is set)
        ;      For each pass:
        ;           a. Assemble group-overlap matrix and vector from group 
        ;              difference table and applicable group weights for the pass.
        ;           b. Solve matrix for group offsets
        ;   5. Apply net group offsets to BCD difference table.
        ; *************************************************************************


        if (keyword_set(GROUP_TAGS)) then begin

	    print, mess + !stime + ': Calculating group offsets.
	    print, mess + !stime + ':  ' + strtrim(string(ngroups),2) + ' total groups.'


            ; preprocess for within-group outliers
            ; note that BCDs deweighted here are "permanently" deweighted
            n_deweight_group = lonarr(ngroups)
            if (keyword_set(GROUP_OUTLIER_THRESH)) then begin
                ; step through difference table rows
                for i = 0L,n-1L do begin
                    istart = ija[i] - 1L  ; index range for off-diagonals in row.
                    istop = ija[i+1] - 2L
                    if (istart le istop) then begin 
                        ; there are some off-diagonals in this row.
                        sa_this = sa[istart:istop]
                        js = ija[istart:istop] - 1L ; count from 0
                        gi = group_index[master_ptr[i]] 
                        gj = group_index[master_ptr[js]]
                        w = where(gj eq gi,wc) ; consider only within-group overlaps
                        if (wc gt 0) then begin
                            js = js[w]
                            sa_this = sa_this[w]
                            med = median(abs(sa_this))
                            if (med ge group_outlier_thresh) then begin
                                ; weights[master_ptr[i]] = 0B
                                master_weights[master_ptr[i]] = 0B ; fix 21SEP2007 DRM
                                n_deweight_group[gi] = n_deweight_group[gi] + 1
                                n_deweighted = n_deweighted + 1L
                                deweighted[master_ptr[i]] = 1B
                            endif
                        endif
                    endif
                endfor
            endif

            if (keyword_set(INTRA_GROUP)) then begin

                ; Do a within-group BCD to BCD overlap correction, to avoid 
                ; local large deviations in between-group comparisons.
                ; Basic idea: assemble "final" matrix and vector from the difference table 
                ;    for solution, but include only within-group overlaps.
                ; The resulting offsets are saved only to apply to the group-to-group matching.
                ; Hard-wired for pretty stiff solution, really don't want any spurious trending

                print, mess + !stime + ': Doing within-group BCD matching.'
                intra_stiffness = .1 
                sa_save = sa ; want to recover original difference table afterward.
                ija_save = ija
                top_index = n + 1L ; initialize end-of-array pointer

                ; assemble matrix/vector from difference table, include only within-group overlaps.
                ; see comments below where the difference-table-to-matrix conversion is done for final result.
                ; step through difference table rows

                for i = 0L,n-1L do begin
                    ; extract off-diagonal index range for this row
                    istart = ija[i] - 1L  ; ija is count-from-1
                    istop = ija[i+1] - 2L
                    ; wgti = weights[master_ptr[i]]
                    wgti = master_weights[master_ptr[i]] ; fix 21SEP2007 DRM
                    if (wgti eq 2B) then wgti = 0B
                    ; set final row-start pointer for this row
                    ija[i] = top_index + 1L ; regardless whether any off-diagonals or not
                    sa[i] = 1. ; default diagonal value if no overlaps, force correction to 0.
                    if (istart le istop) then begin
                        ; there are off-diagonal elements in this row.
                        sa_this = sa[istart:istop]
                        ija_this = ija[istart:istop] ; N.B. keep cols as count from 1
                        gi = group_index[master_ptr[i]]
                        gj = group_index[master_ptr[ija_this-1L]] ; convert col indices to count-from-0
                        w = where(gj eq gi,wc) ; include only same-group overlaps
                        if (wc gt 0) then begin
                            sa_this = sa_this[w]
                            ija_this = ija_this[w]
                            if (wgti eq 0B) then begin
                                ; use all (within-group) overlaps to influence weight=0 BCD
                                n_sa = n_elements(sa_this)
                                ; diagonal for this row 
                                sa[i] = float(n_sa) 
                                sa[top_index:top_index+n_sa-1] = -1.        ; off-diagonal values
                                ija[top_index:top_index+n_sa-1] = ija_this  ; off diagonal columns
                                vec[i] = total(sa_this)                     ; RHS value
                                top_index = top_index + n_sa
                            end else begin
                                ; use only overlaps with weight=1 BCDs to influence weight=1 BCD
                                ; wgts = weights[master_ptr[ija_this-1L]]
                                wgts = master_weights[master_ptr[ija_this-1L]] ; fix 21SEP2007 DRM
                                w = where(wgts eq 2B,wc)
                                if (wc gt 0) then wgts[w] = 0B
                                w = where(wgts eq 1B,wc)
                                if (wc gt 0) then begin
                                    sa[i] = float(wc)*(1.d + intra_stiffness)
                                    sa[top_index:top_index+wc-1] = -1.
                                    ija[top_index:top_index+wc-1] = ija_this[w]
                                    vec[i] = total(sa_this[w])
                                    top_index = top_index + wc
                                endif ; else this is weight=1 BCD with no overlaps, no off_diagonals included.
                            endelse
                        endif
                    endif 
                endfor

                ija[n] = top_index + 1L ; flag for size of sa/ija arrays

                ; solve matrix/vector for offset solutions
                again:
                if (not keyword_set(external) and not keyword_set(internal)) then begin
                    ; use built-in linbcg routine (array size limitations).
                    print, mess + !stime + ': Starting linbcg.'
                    m_sparse = {sa:dblarr(top_index), ija:lon64arr(top_index)}
                    for ii = 0L,top_index-1L do begin ; explicit loop to save memory
                        m_sparse.sa[ii] = double(sa[ii])
                        m_sparse.ija[ii] = long64(ija[ii])
                    endfor
                    intragroup_offsets = linbcg(m_sparse, vec, replicate(0.d,n), /double, iter=n_iter)
                    ; if (min(intragroup_offsets) eq 0. and max(intragroup_offsets) eq 0.) then begin
                    if (max(abs(intragroup_offsets)) le 1.d-8) then begin  ; 12DEC2006 DRM
                        print, mess + !stime + ': Possible problem in linbcd, no offsets determined.'
                    end else begin
                        print, '      linbcg: ', n_iter, ' iterations.', format='(a,i5,a)'
                    endelse
                    m_sparse = 0
                end else if (keyword_set(internal)) then begin
                    ; use linbcg function coded in IDL (no arbitrary limitations but slower)
                    print, mess + !stime + ': Starting linbcg_idl.'
                    intragroup_offsets = linbcg_idl2(sa, ija, vec, replicate(0.d,n), iter=n_iter)
                    print, mess + !stime + ': linbcg_idl2 finished.'
                    print, '      linbcg_idl2: ', n_iter, ' iterations.', format='(a,i5,a)'
                end else if (keyword_set(external)) then begin
                    ; use external compiled function
                    if (n_elements(byte(external)) gt 1) then prog = external else prog = 'linbcg.x'
                    prog_list = findfile(prog,count=c)
                    if (c eq 0) then begin
                        print, mess + !stime + ': linbcg executable not found, changing to internal linbcg.'
                        external = 0 & internal = 1
                        goto, again
                    endif
                    mat_file = 'linbcg_temp.mat'
                    print, mess + !stime + ': Writing matrix file ' + mat_file
                    openw, ilun, mat_file, /get_lun
                    writeu, ilun, long(n), long(n_elements(sa)), vec, sa, ija
                    close, ilun & free_lun, ilun
                    command = prog + ' ' + mat_file + ' linbcg_temp.sol'
                    print, mess + !stime + ': Executing ' + prog
                    spawn, command
                    openr, ilun, 'linbcg_temp.sol', /get_lun
                    s1 = lonarr(1)
                    intragroup_offsets = fltarr(n)
                    readu, ilun, s1, intragroup_offsets
                    close, ilun & free_lun, ilun
                endif

                ; revert to original difference table.
                sa = temporary(sa_save)
                ija = temporary(ija_save)

            end else intragroup_offsets = fltarr(n) ; end INTRA_GROUP option.


            ; assemble overlap difference table for group corrections
            ; group weights ignored here, compile all overlaps

            group_sumdiff = dblarr(ngroups,ngroups)
            group_ndiff = lonarr(ngroups,ngroups)
            
            ; group_sumdiff[i,j] = sum of overlap differences between group i BCDs and group j BCDs
            ; group_ndiff[i,j] = number of group i to group j BCD overlaps

            ; Basic idea: Step through all BCD difference table elements, and update the
            ;   group difference table from the group tags of the overlapping BCDs. 
            ;   Include only those overlap differences that are between different groups.
            ;   Ignore weight=0B BCDs.

            for i = 0L,n-1L do begin
                istart = ija[i] - 1L
                istop = ija[i+1] - 2L
                ; wgti = weights[master_ptr[i]]
                wgti = master_weights[master_ptr[i]] ; fix 21SEP2007 DRM
                if (istart le istop and wgti eq 1B) then begin  ; ignore weight=0B BCDs
                    ; some off-diagonal elements in row.
                    sa_this = sa[istart:istop]
                    js = ija[istart:istop] - 1L ; count from 0
                    ; wgtj = weights[master_ptr[js]]
                    wgtj = master_weights[master_ptr[js]] ; fix 21SEP2007 DRM
                    gi = group_index[master_ptr[i]]
                    gj = group_index[master_ptr[js]]
                    ; consider only intergroup overlaps.
                    w = where(gj ne gi and wgtj eq 1B,wc) ; and ignore weight=0B overlap BCDs
                    if (wc gt 0) then begin
                        ; some intergroup overlaps for this BCD.
                        gj = gj[w]
                        sa_this = sa_this[w]
                        js = js[w]
                        for k = 0L,wc-1L do begin
                            correction = -intragroup_offsets[i] + intragroup_offsets[js[k]]
                            group_sumdiff[gj[k],gi] = group_sumdiff[gj[k],gi] + sa_this[k] + correction
                            group_ndiff[gj[k],gi] = group_ndiff[gj[k],gi] + 1L 
                        endfor
                    endif
                endif
            endfor


            ; determine number of group correction passes
            if (keyword_set(GROUP_ANCHORS) or keyword_set(GROUP_ONEPASS)) then begin
                ; one pass with specified stiffness.
                start_pass = 1
                stop_pass = 1
            end else if (keyword_set(GROUP_FLOATS)) then begin
                ; do final pass with specified stiffness and no deweighting
                start_pass = 1
                stop_pass = 2
            end else begin
                ; initial pass for finding outliers with low stiffness, 
                ; 1 pass with specified stiffness and outlier deweighting
                ; final pass with specified stiffness and no deweighting
                start_pass = 0
                stop_pass = 2
            endelse

            ; Pass 0 = "auto-deweighting" pass: Determine group offsets with
            ;    no group deweighting and hardwired low stiffness. One-third
            ;    of the largest-magnitude offsets are deweighted.
            ; Pass 1 = Determine group offsets with specified group deweighting
            ;    and stiffness.
            ; Pass 2 = Apply corrections from Pass 1 and recalculate offsets
            ;    with no deweighting, for final normalization of levels.
            ; Final group offsets are Pass 1 + Pass 2 (if applicable) results.

            for ipass = start_pass,stop_pass do begin

            print, mess + !stime + ': Group correction pass '+strtrim(string(ipass),2)

            if (ipass eq 0) then use_group_stiffness = .01
            if (ipass eq 1 or ipass eq 2) then use_group_stiffness = group_stiffness

            ; assemble matrix and vector from difference table

            group_m = dblarr(ngroups,ngroups)
            group_vec = dblarr(ngroups)

            ; if not GROUP_EQUALIZE:
            ; group_m[i,i] = total number of intergroup BCD overlaps for the group
            ; group_m[i,j] = (minus) total BCD overlaps between group i and group j 
            ; group_vec[i] = total intergroup BCD overlap differences for group i
            ; if GROUP_EQUALIZE:
            ; group_m[i,i] = total number of overlapping groups for group i
            ; group_m[i,j] = -1 for overlapping groups, 0 otherwise
            ; group_vec[i] = total of average BCD overlap for each group overlap

            for i = 0L,ngroups-1L do begin
                this_row = reform(group_sumdiff[*,i]) 
                this_row_n = reform(group_ndiff[*,i])  ; column is 1st index.
                wgti = group_weights[i]
                w = where(this_row_n gt 0,wc)
                if (wc gt 0) then begin
                    ; there are some group to group overlaps for this group.
                    if (keyword_set(GROUP_EQUALIZE)) then begin
                        ; use average group-to-group BCD overlap
                        this_row[w] = this_row[w]/double(this_row_n[w])
                        this_row_n[w] = 1L
                    endif
                    m_row = dblarr(ngroups)
                    if (wgti eq 0B) then begin
                        ; all other groups influence this group.
                        m_row[w] = -double(this_row_n[w])
                        group_m[*,i] = m_row
                        group_m[i,i] = double(total(this_row_n[w]))
                        group_vec[i] = total(this_row[w])
                    end else begin
                        ; only weight=1 groups influence this group.
                        w = where(this_row_n gt 0 and group_weights eq 1B,wc)
                        if (wc gt 0) then begin
                            m_row[w] = -double(this_row_n[w])
                            group_m[*,i] = m_row
                            group_m[i,i] = double(total(this_row_n[w]))
                            group_vec[i] = total(this_row[w])
                        endif
                    endelse
                end else begin
                    group_m[i,i] = 1.d
                    group_vec[i] = 0.d
                endelse
            endfor
                        
            ; set stiffness, set diagonal for non-overlapping groups to 1 (force correction to 0).
            for i = 0L,ngroups-1L do begin
                ; if (group_m[i,i] eq 0.) then group_m[i,i] = 1.d
                if (abs(group_m[i,i]) lt 1.e-6) then group_m[i,i] = 1.d ; 12DEC2006 DRM
                if (group_weights[i] eq 1B) then group_m[i,i] = group_m[i,i] * (1.d + use_group_stiffness)
            endfor

            ; solve for group offsets.
            if (group_stiffness gt 1.e-6) then begin
                ; use non-singular matrix solver
                ludc, group_m, index
                group_offsets = lusol(group_m, index, group_vec, /DOUBLE)
            end else begin
                ; use singular matrix solver
                svdc, group_m, w, u, v, /DOUBLE
                wmax = max(w)
                wptr = where(w/wmax lt 1.0D-6, wcount)
                nw = temporary(w)
                if (wcount gt 0) then nw[wptr] = 0.0
                group_offsets = svsol(u, nw, v, group_vec, /DOUBLE) 
            endelse
            
            if (ipass eq 0) then begin
                ; auto deweighting of groups, deweight quarter with largest corrections for now.
                aoff = abs(group_offsets)
                ptr = sort(aoff)
                max_index = round(float(ngroups)*.75)
                max_offset = aoff(ptr(max_index))
                w = where(aoff ge max_offset,wc)
                if (wc gt 0) then group_weights[w] = 0B
            end else if (ipass eq 1) then begin
                net_group_offsets = group_offsets
            end else if (ipass eq 2) then begin
                net_group_offsets = net_group_offsets + group_offsets
            endif
          
            if (ipass gt 0) then begin
                if (keyword_set(GROUP_AOR))  then print, '         AOR      CORRECTION   WGT   BCDs DEWEIGHTED' else $
                                                  print, '         TAG      CORRECTION   WGT   BCDs DEWEIGHTED' 
                for i = 0L,ngroups-1L do begin
                    print, '    ', strtrim(string(group_taglist[i]),2), $
                        -net_group_offsets[i], group_weights[i], n_deweight_group[i], $
                        format='(a,a10,2x,f9.3,2x,i6,3x,i4)'
                endfor
            endif

            if (ipass eq 1 and stop_pass eq 2) then begin
                ; apply corrections to group difference table
                for i = 0L,ngroups-1L do begin
                for j = 0L,ngroups-1L do begin
                    if (group_ndiff[j,i] gt 0) then begin
                        xn = double(group_ndiff[j,i])
                        group_sumdiff[j,i] = group_sumdiff[j,i] - xn*(group_offsets[i] - group_offsets[j])
                    endif
                endfor
                endfor
                group_weights[*] = 1B ; no deweighted groups for 2nd pass.
            endif


            endfor ; loop through group offset solution passes


            ; apply final group offsets to BCD difference table.
            g_offsets = fltarr(n) ; store group offsets for each BCD, in input order
            for i = 0L,n-1L do begin
                g_offsets[master_ptr[i]] = net_group_offsets[group_index[master_ptr[i]]] 
                istart = ija[i] - 1L
                istop = ija[i+1] - 2L
                if (istart le istop) then begin
                    ; some off-diagonal elements in row.
                    sa_this = sa[istart:istop]
                    js = ija[istart:istop] - 1L
                    gi = group_index[master_ptr[i]]
                    gj = group_index[master_ptr[js]]
                    sa_this = sa_this - net_group_offsets[gi] + net_group_offsets[gj]
                    sa[istart:istop] = sa_this
                endif
            endfor


        endif ; group correction option
 


        ; *************************************************************************
        ; Now that group corrections are complete, can apply group deweighting
        ; and/or aor deweighting if specified.
        ; *************************************************************************

        if (keyword_set(GROUP_DEWEIGHT) and keyword_set(GROUP_TAGS)) then begin
            ngrp_dw = 0L
            for i = 0L,n_elements(group_deweight)-1L do begin
                w = where(group_tags eq group_deweight[i],wc)
                ; if (wc gt 0) then weights[w] = 0B
                if (wc gt 0) then master_weights[w] = 0B ; fix 21SEP2007 DRM
                ngrp_dw = ngrp_dw + wc
            endfor
            print, mess + !stime + ': ' + strtrim(string(ngrp_dw),2) + ' BCDs deweighted by group.' 
        endif
       
        if (keyword_set(AOR_DEWEIGHT)) then begin
            naor_dw = 0L
            for i = 0L,n_elements(aor_deweight)-1L do begin
                w = where(master_aor eq aor_deweight[i],wc)
                if (wc gt 0) then master_weights[w] = 0B 
                naor_dw = naor_dw + wc
            endfor
            print, mess + !stime + ': ' + strtrim(string(naor_dw),2) + ' BCDs deweighted by AOR.' 
        endif


        ; *************************************************************************
        ; Now find global overlap solution
        ; *************************************************************************



        ; *************************************************************************
        ; Convert difference table to LHS matrix and make RHS vector
        ; LHS : off-diagonals = -1, diagonals = number of overlaps + stiffness factor
        ; RHS = total of overlap differences in row
        ; Matrix and vector values are adjusted by overlap weights if applied.
        ; Also, "deweighted" BCDs cause some overlaps to be deleted from result.
        ;
        ; N.B. The sparse matrix may be reduced in size from difference table 
        ; due to deweighted BCDs. This can be done "in-place" since we are always
        ; dereferencing the original arrays at higher indices than we are
        ; overwriting.
        ; *************************************************************************

        deweighted_stiffness = 0. ; may test very small values instead later

        top_index = n + 1L

        for i = 0L,n-1L do begin
            istart = ija[i] - 1L  ; ija is count-from-1
            istop = ija[i+1] - 2L
            ; wgti = weights[master_ptr[i]]
            wgti = master_weights[master_ptr[i]] ; fix 21SEP2007 DRM
            if (wgti eq 2B) then begin 
                ; defunct, I think.
                if (iter eq 0) then wgti = 0B else wgti = 1B
            endif
            ; set final row-start pointer for this row
            ija[i] = top_index + 1L ; regardless whether any off-diagonals or not
            sa[i] = 1. ; default diagonal value if no overlaps, force correction to 0.
            if (istart le istop) then begin
                ; there are off-diagonal elements in this row.
                sa_this = sa[istart:istop]
                ija_this = ija[istart:istop]
                ; if (keyword_set(weight_overlaps)) then $
                if (keyword_set(wa)) then $
                    ; weighting by overlap area.
                    wa_this = float(wa[istart:istop])/127. else wa_this = 1.
                if (wgti eq 0B) then begin
                    ; use all overlaps to influence weight=0 BCD
                    n_sa = n_elements(sa_this)
                    ; diagonal for this row 
                    ; sa[i] = total(wa_this*(fltarr(n_sa)+1.)) 
                    sa[i] = total(wa_this*(fltarr(n_sa)+1.) + float(n_sa)*deweighted_stiffness)
                    sa[top_index:top_index+n_sa-1] = -1. * wa_this   ; off-diagonal values
                    ija[top_index:top_index+n_sa-1] = ija_this       ; off diagonal columns
                    vec[i] = total(sa_this*wa_this)                  ; RHS value
                    overlaps[master_ptr[i]] = n_sa
                    top_index = top_index + n_sa
                end else begin
                    ; use only overlaps with weight=1 BCDs to influence weight=1 BCD
                    ; wgts = weights[master_ptr[ija_this-1L]]
                    wgts = master_weights[master_ptr[ija_this-1L]] ; fix 21SEP2007 DRM
                    w = where(wgts eq 2B,wc)
                    if (wc gt 0) then if (iter eq 0) then wgts[w] = 0B else wgts[w] = 1B
                    w = where(wgts eq 1B,wc)
                    if (wc gt 0) then begin
                        sa[i] = total(wa_this[w]*(fltarr(wc)+1.)) + float(wc)*stiffness[0]
                        sa[top_index:top_index+wc-1] = -1. * wa_this[w]
                        ija[top_index:top_index+wc-1] = ija_this[w]
                        vec[i] = total(sa_this[w]*wa_this[w])
                        overlaps[master_ptr[i]] = wc
                        top_index = top_index + wc
                    endif ; else this is weight=1 BCD with no overlaps, no off_diagonals included.
                endelse
            endif 
        endfor

        ija[n] = top_index + 1L
 
        ; free some memory
        ; if (keyword_set(weight_overlaps)) then wa = 0
        ; if (keyword_set(wa)) then wa = 0
        print, 'Saving wa array.'

        ; N.B. At this point the physical arrays sa and ija are likely bigger than
        ; the number of valid elements. There's no way to trim them down to
        ; true size without allocating additional memory, so need to determine if
        ; solving routines expect true-size arrays or not.



        ; *************************************************************************
        ; Loop through solution iterations
        ; *************************************************************************

        ; keep record of iteration-by-iteration corrections
        offsets_i = fltarr(n_iterations,n)

        for iter_i = 0,n_iterations-1 do begin


        ; *************************************************************************
        ; Turn sa, ija arrays into sparse structure if necessary (eats memory)
        ; *************************************************************************

        if (keyword_set(external)) then begin
            mat_file = 'linbcg_temp.mat'
            print, mess + !stime + ': Writing matrix file ' + mat_file
            openw, ilun, mat_file, /get_lun
            writeu, ilun, long(n), long(n_elements(sa)), vec, sa, ija
            close, ilun & free_lun, ilun
            print, '    n    = ' + strtrim(string(n),2)
            print, '    size = ' + strtrim(string(n_elements(sa)),2)
            if (iter_i eq (n_iterations-1)) then begin
                ; free memory after last iteration
                sa = 0
                ija = 0
            endif
        endif

        if (not keyword_set(external) and not keyword_set(internal)) then begin
            ; Make ija = long64 (8-byte) version for input to linbcg
            ; also for conversion to full NxN matrix for SVD input if necessary.
            ; I think big memory drain -- this usage of temporary doesn't seem
            ; to work as you might think. Requires at least 1.5X additional memory.
            m_sparse = {sa:dblarr(top_index), ija:lon64arr(top_index)}
            for ii = 0L,top_index-1L do begin ; explicit loop to save memory
                m_sparse.sa[ii] = double(sa[ii])
                m_sparse.ija[ii] = long64(ija[ii])
            endfor
        endif


        ; *************************************************************************
        ; Solve for offsets
        ; *************************************************************************

        ; stop

        ; Use sparse matrix operations if applicable (stiffness sufficiently high)
	print, mess + !stime + ': Solving for offsets'
      
        if (use_svd eq 0) then begin
           ; sparse matrix solution
	   print, mess + !stime + ': Using sparse matrix operations.'
           if (not keyword_set(external) and not keyword_set(internal)) then begin
               ; default, use built-in IDL function
               offsets = linbcg(m_sparse, vec, replicate(0.d,n), /double, iter=n_iter)
               print, '      linbcg: ', n_iter, ' iterations.', format='(a,i5,a)'
               ; if (min(offsets) eq 0. and max(offsets) eq 0.) then begin
               if (max(abs(offsets)) lt 1.e-6) then begin ; 12DEC2006 DRM
                   print, mess + !stime + ': Possible problem in linbcd, no offsets determined.'
                   print, mess + !stime + ': Terminating without applying offsets.'
                   goto, skip_solution 
                endif
           end else if (keyword_set(internal)) then begin
               ; use function coded in IDL
               print, mess + !stime + ': Starting linbcg_idl.'
               offsets = linbcg_idl2(sa, ija, vec, replicate(0.d,n), iter=n_iter)
               print, mess + !stime + ': linbcg_idl2 finished.'
               print, '      linbcg_idl2: ', n_iter, ' iterations.', format='(a,i5,a)'
           end else if (keyword_set(external)) then begin
               ; use external compiled function
               if (n_elements(byte(external)) gt 1) then prog = external else prog = 'linbcg.x'
               prog_list = findfile(prog,count=c)
               if (c eq 0) then begin
                  print, mess + !stime + ': linbcg executable not found, exiting.'
                  return
               endif
               command = prog + ' ' + mat_file + ' linbcg_temp.sol'
               print, mess + !stime + ': Executing ' + prog
               spawn, command
               openr, ilun, 'linbcg_temp.sol', /get_lun
               s1 = lonarr(1)
               offsets = fltarr(n)
               readu, ilun, s1, offsets
               close, ilun & free_lun, ilun
           endif
           print, mess + !stime + ': No global normalization applied.'
           m_sparse = 0
        end else begin
           ; SVD solution
	   print, mess + !stime + ': Using SVD matrix operations.'
	   print, mess + !stime + ': Converting sparse matrix to conventional representation.'
           ; recover full matrix array from sparse representation
           m_full = fulstr(temporary(m_sparse))
           ; stop
           ; m_full[*,n-1] = 1. ; apply normalization right in matrix
           ; vec[n-1] = 0.
               svdc, m_full, w, u, v, /DOUBLE
               ; Find almost singular values
               wmax = max(w)
               wptr = where(w/wmax lt 1.0D-8, wcount)
               wptr1 = where(w/wmax lt 1.0D-3, wcount1)
               wptr2 = where(w/wmax lt 1.0D-2, wcount2)
               wptr3 = where(w/wmax lt 1.0D-1, wcount3)
               nw = temporary(w)
               ; and set their 'weights' to zero
               print, wcount, ' elements zeroed. (1.D-6)'
               if (wcount gt 0) then print, nw[wptr]/wmax
               print, wcount1, ' elements zeroed. (1.D-3)'
               if (wcount1 gt 0) then print, nw[wptr1]/wmax
               print, wcount2, ' elements zeroed. (1.D-2)'
               if (wcount2 gt 0) then print, nw[wptr2]/wmax
               print, wcount3, ' elements zeroed. (1.D-1)'
               if (wcount3 gt 0) then print, nw[wptr3]/wmax
               if (wcount gt 0) then nw[wptr] = 0.0
               offsets = svsol(u, nw, v, vec, /DOUBLE)
           ; ludc, m_full, index
           ; offsets = lusol(m_full, index, vec)
           m_full = 0 & nw = 0 & v = 0 
        endelse

        if (use_svd eq 1) then begin
           print, mess + !stime + ': Normalize for offsets'
           ; global = total(float(weights)*offsets) / float(ngood)
           global = total(float(master_weights)*offsets) / float(ngood) ; 21SEP2007 DRM
           offsets = offsets - global
           if (n_iterations gt 1) then begin
               print, 'Multiple iterations pointless for SVD solution.'
               goto, skip_iterations
           endif
        endif

        offsets_i[iter_i,*] = offsets

        if (iter_i lt (n_iterations-1)) then begin
            ; apply corrections to RHS difference vector to prepare next iteration.
            for i = 0L,n-1L do begin
                istart = ija[i] - 1L  ; ija is count-from-1
                istop = ija[i+1] - 2L
                if (istart le istop) then begin
                    n_off = istop - istart + 1L ; number of off-diagonal elements
                    wgts = abs(sa[istart:istop]) ; sa contains -wgt or -1. as values.
                    indices = ija[istart:istop] - 1L ; to count-from-0
                    vec[i] = vec[i] - total(wgts*((fltarr(n_off) + offsets[i]) - offsets[indices]))
                endif
            endfor
        endif
                    

        endfor ; loop over correction iterations


        offsets = total(offsets_i,1)
        
        ; write out offsets as a FITS file.
        if (keyword_set(MAKE_OFFSETS_FILE)) then begin
            st = !stime
            day = strtrim(strmid(st,0,2),2)
            if (strlen(day) eq 1) then day = '0' + day
            month = strupcase(strmid(st,3,3))
            year = strmid(st,7,4)
            hour = strmid(st,12,2)
            minute = strmid(st,15,2)
            second = strmid(st,18,2)
            writefits, 'offsets_'+day+month+year+'_'+hour+minute+second+'.fits', offsets_i
        endif

        skip_iterations:

        offsets_i = 0


        ; *************************************************************************
        ; reorder the offsets to true input list order
        ; *************************************************************************

        temp1 = fltarr(n)
        temp1[master_ptr] = offsets
        offsets = temp1

        ; *************************************************************************
        ; add group corrections if necessary to get total corrections
        ; g_offsets is already in input BCD order.
        ; Note that g_offsets have already been applied to original difference 
        ; table -- here we are compiling the net offset correction for each BCD.
        ; *************************************************************************

        if (keyword_set(g_offsets)) then offsets = offsets + g_offsets

        ; *************************************************************************
        ; restore BCD and mask files lists to input value if necessary
        ; *************************************************************************

        if (keyword_set(infiles_save)) then infiles = temporary(infiles_save)
        if (keyword_set(mask_files_save)) then mask_files = temporary(mask_files_save)

        ; *************************************************************************
        ; Make output diagnostics structure if necessary
        ; *************************************************************************

        if (arg_present(diagnostics)) then begin
            diagnostics = {offsets:offsets, overlaps:overlaps, $
                median_deviations:median_deviations, mean_deviations:mean_deviations, $
                deweighted:deweighted, vector:vec[master_ptr], full_table:full_table}
        endif

        ; *************************************************************************
        ; Now apply offsets to the data
        ; N.B. offsets array is now in original BCD order (augmented by pertitions)
        ; *************************************************************************

	print, mess + !stime + ': Apply offsets and write corrected BCDs'
        if (keyword_set(outtext)) then openw, ilun, outtext, /get_lun

        k = 0L ; running index in partitioned list.

        ; step through in original (unpartitioned) BCD order.

        for i = 0L,n_save-1L do begin

            k_start = k
            iim = readfits(infiles[i], ih, /SILENT)
            ; See if this BCD has multiple partitions
            if (entries[i] eq 1) then begin
                ; Apply offset to whole array
                ptr = where(iim eq iim,ptrc)
                if (ptrc gt 0) then iim[ptr] = iim[ptr] - offsets[k]
                sxaddpar, ih, 'OFFSET0', -offsets[k], 'Additive offset applied.'
                sxaddpar, ih, 'LOROWO', master_rowlo[k], 'Lower row for segment.'
                sxaddpar, ih, 'HIROWO', master_rowhi[k], 'Upper row for segment.'
                sxaddpar, ih, 'NOVREG0', overlaps[k], 'Number of overlapping segments.'
                sxaddpar, ih, 'MEDDEV0', median_deviations[k], 'Median of absolute differences.'
                sxaddpar, ih, 'MEANDEV0', mean_deviations[k], 'Mean of absolute differences.'
                if (keyword_set(deweighted)) then sxaddpar, ih, 'DEWGT0', fix(deweighted[k]),'0=normal, 1=deweighted'
                if (keyword_set(outtext)) then printf, ilun, infiles[i], 0, master_rowlo[k], master_rowhi[k], $
                    overlaps[k], -offsets[k], format='(a,1x,i2,1x,i3,1x,i3,1x,i4,1x,f7.1)'
                k = k + 1L
            end else begin
                for j = 0L,entries[i]-1 do begin
                    ; Apply offset to partition region
                    rowlo = master_rowlo[k]
                    rowhi = master_rowhi[k]
                    iim_seg = iim[*,rowlo:rowhi]
                    ptr = where(iim_seg eq iim_seg,ptrc)
                    if (ptrc gt 0) then iim_seg[ptr] = iim_seg[ptr] - offsets[k]
                    iim[*,rowlo:rowhi] = iim_seg
                    if (rowlo gt 0 and master_sourcecol[k] ge 0) then begin
                        ; (master_sourcecol[k] is -999 if filtering is to be suppressed.)
                        spike_filter, iim, rowlo, iim_out, column=master_sourcecol[k], spike_halfwin=spike_halfwin
                        iim = iim_out
                    endif
                    jstr = strtrim(string(j),2)
                    sxaddpar, ih, 'OFFSET'+jstr, -offsets[k], 'Additive offset applied.'
                    sxaddpar, ih, 'LOROW'+jstr, master_rowlo[k], 'Lower row for segment.'
                    sxaddpar, ih, 'HIROW'+jstr, master_rowhi[k], 'Upper row for segment.'
                    sxaddpar, ih, 'NOVREG'+jstr, overlaps[k], 'Number of overlapping segments.'
                    sxaddpar, ih, 'MEDDEV'+jstr, median_deviations[k], 'Median of absolute differences.'
                    sxaddpar, ih, 'MEANDEV'+jstr, mean_deviations[k], 'Mean of absolute differences.'
                    if (keyword_set(deweighted)) then sxaddpar, ih, 'DEWGT'+jstr, fix(deweighted[k]), '0=normal, 1=deweighted' 
                    if (keyword_set(outtext)) then printf, ilun, infiles[i], j, master_rowlo[k], master_rowhi[k], $
                        overlaps[k], -offsets[k], format='(a,1x,i2,1x,i3,1x,i3,1x,i4,1x,f7.1)'
                    k = k + 1L
                endfor
            endelse
            if (keyword_set(outfiles)) then begin
                if (keyword_set(out_dir)) then begin
                    bare = strip_dir(outfiles[i])
                    outfile = out_dir+'/'+bare
                    file_exists = file_test(outfile)
                    if (file_exists) then begin
                        print, 'No overwrite for BCD file: ', outfile
                    end else begin
                        writefits, outfile, iim, ih
                    endelse
                end else writefits, outfiles[i], iim, ih
            end else if (keyword_set(out_dir)) then begin
                ; out_dir specified alone
                bare = strip_dir(infiles[i])
                outfile = out_dir+'/'+bare
                file_exists = file_test(outfile)
                if (file_exists) then begin
                    ; print, 'No overwrite for BCD file: ', outfile
                    ; print, 'Overwriting BCD file: ', outfile
                    ; allow overwriting for now...
                    writefits, outfile, iim, ih
                end else begin
                    writefits, outfile, iim, ih
                endelse
            endif
            corrections[i] = -offsets[k_start]
            if (k_start ge i) then begin
                ; pack 1st segment stats into output arrays
                ; (something reasonable for now...)
                offsets[i] = offsets[k_start]
                overlaps[i] = overlaps[k_start]
                deweighted[i] = deweighted[k_start]
                median_deviations[i] = median_deviations[k_start]
                mean_deviations[i] = mean_deviations[k_start]
            endif

            ; print progress status at intervals
            if (i lt 1000L) then if ((i mod 100L) eq 0 and i gt 0) then $
                print, i, ' BCDs written', format='(i6,a)'
            if (i ge 1000L and i lt 5000L) then if ((i mod 500L) eq 0) then $
                print, '   ', i, ' BCDs written', format='(a,i7,a)'
            if (i ge 5000L and i lt 20000L) then if ((i mod 1000L) eq 0) then $
                print, '   ', i, ' BCDs written', format='(a,i7,a)'
            if (i ge 20000L) then if ((i mod 5000L) eq 0) then $
                print, '   ', i, ' BCDs written', format='(a,i7,a)'

        endfor

        if (k ne n) then begin
            ; some kind of accounting error...
        endif
                
        if (k gt n_save) then begin
            ; cut down output arrays to true lengths.
            offsets = offsets[0:n_save-1]
            corrections = corrections[0:n_save-1]
            overlaps = overlaps[0:n_save-1]
            deweighted = deweighted[0:n_save-1]
            median_deviations = median_deviations[0:n_save-1]
            mean_deviations = mean_deviations[0:n_save-1]
        endif

        if (keyword_set(outtext)) then begin
            close, ilun
            free_lun, ilun
        endif

        skip_solution:

        print
        print, mess + !stime + ': Processing concluded.'
        print

return
end
