pro apply_q, ra1, dec1, q, ra2, dec2, goo1, goo2, INVERSE=inverse
;+
; NAME:
; APPLY_Q
; PURPOSE:
; Apply quaternion (rotation) to a source coordinate pair
;
; CALLING SEQUENCE:
; APPLY_Q, ra, dec, q, nra, ndec
;
; INPUTS:
; RA =  Source Right Ascension in degrees (double scalar/vector)
; DEC = Source Declination in degrees
; Q = Quaternion to apply (double (*,4)), first three components are 
;     vector [x, y, z], 4th component is scalar, if q is 4 element vector,
;     then that quaternion is applied to all ra, dec pairs
; 
; KEYWORD:
; INVERSE if set, perform the inverse rotation; that is, substitute inverse
; q for q, inverse(q) = [-q1, -q2, -q3, q4], code handles inverse rotations
; by transposing rotation matrix
;
; OUTPUTS
; NRA = Revised Right Ascension after application of the URM
; NDEC = revised Declination
;
; NOTES: 
; Applies rotation specified by input quaternion to input angular
; coordinates. Input coordinates (a, d) are converted to ECI (Cartesian vector)
; reperesentation (position on a unit sphere can be described as a cartesian
; vector of length one) v = [x, y, z].  A rotation matrix is formed from the 
; quaternion:
;     [ q1^2-q2^2-q3^2+q4*^2  2 * (q1*q2 + q3*q4)   2 * (q1*q3 - q2*q4)  ]
; R = | 2 * (q1*q2 - q3*q4)   -q1^2+q2^2-q3^2+q4^2  2 * (q2*q3 + q1*q4)  |
;     [ 2 * (q1*q3 +q2*q4)    2 * (q2*q3 - q1*q4)   -q1^2-q2^2+q3^2+q4^2 ]
; The rotation is applied v' = R * v and the output coordinates are converted
; back to spherical representation
;
; MODIFICATION HISTORY
; Created RFS - Sean J. Carey October, 1998
; Added application of constant quaternion to ra, dec vectors, SJC 01/26/2000
;-

	if (N_params() lt 5) then begin
		print, 'Syntax -- APPLY_Q, ra1, dec1, q, ra2, dec2, /INVERSE'
		return
	endif
; Break quaternion into components
	if (n_elements(ra1) eq 1) then begin
		q1 = q(0)
		q2 = q(1)
		q3 = q(2)
		q4 = q(3)
	endif else begin
		if (n_elements(q) eq 4) then begin
			q1 = replicate(q(0), n_elements(ra1))
			q2 = replicate(q(1), n_elements(ra1))
			q3 = replicate(q(2), n_elements(ra1))
			q4 = replicate(q(3), n_elements(ra1))
		endif else begin
			q1 = q(*, 0)
			q2 = q(*, 1)
			q3 = q(*, 2)
			q4 = q(*, 3)
		endelse
	endelse

	DEG_TO_RAD = !DPI / 180.0D

; Convert celestial coordinates to ECI
	x1 = cos(DEG_TO_RAD * ra1) * cos(DEG_TO_RAD * dec1)
	y1 = sin(DEG_TO_RAD * ra1) * cos(DEG_TO_RAD * dec1)
	z1 = sin(DEG_TO_RAD * dec1)

; Create rotation matrix
	r11 = q1 * q1 - q2 * q2 - q3 * q3 + q4 * q4
	r12 = 2.0 * (q1 * q2 + q3 * q4)
	r13 = 2.0 * (q1 * q3 - q2 * q4)
	r21 = 2.0 * (q1 * q2 - q3 * q4)
	r22 = -q1 * q1 + q2 * q2 - q3 * q3 + q4 * q4
	r23 = 2.0 * (q2 * q3 + q1 * q4)
	r31 = 2.0 * (q1 * q3 + q2 * q4)
	r32 = 2.0 * (q2 * q3 - q1 * q4)
	r33 = -q1 * q1 - q2 * q2 + q3 * q3 + q4 * q4

	if (keyword_set(INVERSE)) then begin
		goo = r12 & r12 = r21 & r21 = goo
		goo = r13 & r13 = r31 & r31 = goo
		goo = r23 & r23 = r32 & r32 = goo
	endif

; rotate coordinates -- Euler rotation of vector
	x2 = r11 * x1 + r12 * y1 + r13 * z1 
	y2 = r21 * x1 + r22 * y1 + r23 * z1
	z2 = r31 * x1 + r32 * y1 + r33 * z1
	goo1 = [x1, y1, z1]
	goo2 = [x2, y2, z2]

; Convert back from ECI
	ra2 = atan(y2, x2) / DEG_TO_RAD
	dec2 = asin(z2) / DEG_TO_RAD
return
end
