; linbcg_idl2.pro
;
; Same as linbcg_idl except takes sa and ija arrays as separate inputs.
; (avoid having to create actual sparse structure)
;
; sparse matrix solver linbcg written in IDL code. 
; algorithm from Numerical Recipes in C
; DRM conversion to IDL 28 Feb 2006
;
; keywords same as linbcg except /double is assumed.
; result = linbcg_idl(sa,ija,b,x,itol=itol,tol=tol,iter=iter,itmax=itmax)
;
; N.B. sa, ija are themselves count-from-0, but ija
; indices are (standard definition) count-from-1.
; Code here has been converted to count-from-0 arrays.
;


pro sprsax, sa, ija, x, b, n
b = double(sa[0:n-1]) * x[0:n-1]
starts = ija[0L:n-1L] - 1L  ; true indices in sa, ija arrays, count from 0
stops = ija[1L:n] - 1L - 1L  ;
w = where(stops ge starts,wc)
if (wc gt 0) then for i = 0L,wc-1L do begin
    k = w[i] ; true row index
    k1 = starts[k]
    k2 = stops[k]
    b[k] = b[k] + total(double(sa[k1:k2]) * x[ija[k1:k2]-1L])
endfor
    ; b[w[i]] = b[w[i]] + $
    ;     total(sa[starts[w[i]]:stops[w[i]]] * x[ija[starts[w[i]]:stops[w[i]]]-1L])
end


pro sprstx, sa, ija, x, b, n
b = double(sa[0:n-1L]) * x[0:n-1L]
starts = ija[0L:n-1L] - 1L ; true indices in sa,ija arrays
stops = ija[1L:n] - 1L - 1L
w = where(stops ge starts,wc)
if (wc gt 0) then for i = 0L,wc-1L do begin
    k = w[i]
    k1 = starts[k]
    k2 = stops[k]
    b[ija[k1:k2]-1L] = b[ija[k1:k2]-1L] + double(sa[k1:k2])*x[k]
endfor
    ; b[ija[starts[w[i]]:stops[w[i]]]-1L] = b[ija[starts[w[i]]:stops[w[i]]]-1L] + $
    ;     sa[starts[w[i]]:stops[w[i]]] * x[w[i]]
end


pro atimes, sa, ija, n, x, r, itrnsp
if (itrnsp eq 1) then sprstx, sa, ija, x, r, n $
    else sprsax, sa, ija, x, r, n
end


pro asolve, sa, ija, n, b, x, itrnsp
sa1 = sa[0:n-1]
wnon = where(sa1 ne 0.,wcn)
wzer = where(sa1 eq 0.,wcz)
if (wcn gt 0) then x[wnon] = b[wnon] / double(sa[wnon])
if (wcz gt 0) then x[wzer] = b[wzer]
end


function snrm, n, sx, itol
if (itol le 3) then return, sqrt(total(sx*sx)) else return, max(abs(sx))
end


function linbcg_idl2, sa, ija, b, x, itol=itol, tol=tol, itmax=itmax, iter=iter, err=err
eps = 1.e-14
n = n_elements(b)
if (not keyword_set(itol)) then itol = 1
if (not keyword_set(tol)) then tol = 1.e-10
if (not keyword_set(itmax)) then itmax = 1000
p = dblarr(n)
pp = dblarr(n)
r = dblarr(n)
rr = dblarr(n)
z = dblarr(n)
zz = dblarr(n)
iter = 0
atimes, sa, ija, n, double(x), r, 0
r = double(b) - r
rr = r
if (itol eq 1) then begin
    bnrm = snrm(n,b,itol)
    asolve, sa, ija, n, r, z, 0
end else if (itol eq 2) then begin
    asolve, sa, ija, n, b, z, 0
    bnrm = snrm(n,z,itol)
    asolve, sa, ija, n, r, z, 0
end else if (itol eq 3 or itol eq 4) then begin
    asolve, sa, ija, n, b, z, 0
    bnrm = snrm(n,z,itol)
    asolve, sa, ija, n, r, z, 0
    znrm = snrm(n,z,itol)
endif
while (iter le itmax) do begin
    iter = iter + 1
    asolve, sa, ija, n, rr, zz, 1
    bknum = total(z*rr)
    if (iter eq 1) then begin
        p = z
        pp = zz
    end else begin
        bk = bknum/bkden
        p = bk*p + z
        pp = bk*pp + zz
    endelse
    bkden = bknum
    atimes, sa, ija, n, p, z, 0
    akden = total(z*pp)
    ak = bknum/akden
    atimes, sa, ija, n, pp, zz, 1
    x = x + ak*p
    r = r - ak*z
    rr = rr - ak*zz
    asolve, sa, ija, n, r, z, 0
    if (itol eq 1) then begin
        err = snrm(n,r,itol)/bnrm
        ; print, '   bnrm = ', bnrm
        ; print, '   snrm = ', snrm(n,r,itol)
    end else if (itol eq 2) then begin
        err = snrm(n,z,itol)/bnrm
    end else if (itol eq 3 or itol eq 4) then begin
        zm1nrm = znrm
        znrm = snrm(n,z,itol)
        if (abs(zm1nrm-znrm) gt eps*znrm) then begin
            dxnrm = abs(ak) * snrm(n,p,itol)
            err = znrm/abs(zm1nrm-znrm) * dxnrm
        end else begin
            err = znrm/bnrm
            continue
        endelse
        xnrm = snrm(n,x,itol)
        if (err le 0.5d*xnrm) then begin
            err = err / xnrm
        end else begin
            err = znrm/bnrm
            continue
        endelse
    endif
    print, 'iter = ', iter, '  err = ', err
    if (err le tol) then goto, done
endwhile
done:
return, x
end

