pro spike_filter, im, n, imout, column=column, spike_halfwin=spike_halfwin

; PURPOSE
;    Procedure to eliminate "diffraction spike" artifact from 
;    sectioned overlap-corrected BCD images (see 
;    many_bcd_overlap_correct_grp2.pro). Interpolates 
;    background levels across the transition region.
;
;    (May work poorly in regions of very bright structured
;    emission.)
;
; INPUTS
;     im = BCD image array (NaNs for no-data assumed.)
;     n = row# for overlap transition
;
; OUTPUTS
;     imout = filtered image array
;
; KEYWORDS
;     column = column number for bright-source. 
;         If set, filtering will be done separately left-and-right
;         of source.
;
; 25 MAY 2006 DRM
;

if (not keyword_set(spike_halfwin)) then spike_halfwin = [4,8]

hw_in = spike_halfwin[0]
hw_out = spike_halfwin[1]
hw_fix = spike_halfwin[0] - 1
min_rows = hw_out - hw_in + 1

s = size(im)
nsize = s(1) ; assume square images.

x = findgen(nsize) ; for tagging row numbers.

; set row ranges for above-and-below-transition interpolation

; "below" row group
lo1 = n - hw_out
if (lo1 lt 0) then lo1 = 0
hi1 = n - hw_in
if (hi1 lt 0) then hi1 = 0
n1 = hi1 - lo1 + 1

; "above" row group
lo2 = n + hw_in
if (lo2 gt (nsize-1)) then lo2 = nsize - 1
hi2 = n + hw_out
if (hi2 gt (nsize-1)) then hi2 = nsize - 1
n2 = hi2 - lo2 + 1

; set "x-y" arrays for linear fitting.
x1 = x[lo1:hi1] ; row numbers.
x2 = x[lo2:hi2]
y1 = fltarr(n1) ; to be assigned row medians.
y2 = fltarr(n2)

; If any row in the filtered region has a NaN or clipped pixel at 
; a given column, eliminate that column from consideration for all rows.

aa = im[*,x1[0]:x2[1]] ; total filtered region.
w = where(aa eq aa,wc) ; true data values.
if (wc gt 0) then begin
    ; determine clipping value, for now 90th percentile pixel value.
    aa1 = aa[w]
    ptr = sort(aa1)
    iclip = round(.9 * float(wc-1)) ; ignore brightest 10% of pixels
    clip = aa1[ptr[iclip]]
    w = where(aa ge clip,wc)
    if (wc gt 0) then aa[w] = 0./0. ; make clipped pixels NaNs.
endif
; any total containing NaN is a NaN.
rowflag = total(aa,2) ; this is a global column-use flag (omit NaNs)


; set column partitions: left-and-right of source if col specified.
c1 = [0,nsize-1] ; default all columns in one partition.
c2 = 0
margin = 32 ; ignore left-and-right partition if source near edge.
if (keyword_set(column)) then begin
    if (column ge margin and column le (nsize-margin)) then begin
        c1 = [0,column-1]
        c2 = [column,nsize-1]
    endif
endif


imout = im

; process each column-partition

for k = 1,2 do begin

    if (k eq 1) then cc = c1
    if (k eq 2) then cc = c2

    if (keyword_set(cc)) then begin

        ; calculate column medians, use only columns less than global median
        ncols = cc[1] - cc[0] + 1
        col_meds = fltarr(ncols)
        for i = cc[0],cc[1] do begin 
           a = [reform(im[i,lo1:hi1]),reform(im[i,lo2:hi2])]
           w = where(a eq a and a gt -10.,wc)
           if (wc gt 0) then col_meds[i-cc[0]] = median(a[w])
        endfor
        w = where(col_meds gt 0.,wc)
        if (wc gt 2) then begin
            use_med = median(col_meds[w])
        end else use_med = 1.e10 ; basically there's no valid data
        ptr_col = where(col_meds gt 0. and col_meds lt use_med,ptrc)
        ; ptr_col = indgen(ncols) ; test
        if (ptrc lt 3) then begin
            ; bail.
        endif

        ; assign row medians for "below" group
        for i = lo1,hi1 do begin
            a = im[cc[0]:cc[1],i]
            flags = rowflag[cc[0]:cc[1]]
            a = a[ptr_col]
            flags = flags[ptr_col]
            w = where(a eq a and flags eq flags,wc)
            if (wc gt 0) then y1[i-lo1] = median(a[w]) else y1[i-lo1] = -999.
        endfor

        ; assign row medians for "above" group
        for i = lo2,hi2 do begin
            a = im[cc[0]:cc[1],i]
            flags = rowflag[cc[0]:cc[1]]
            a = a[ptr_col]
            flags = flags[ptr_col]
            w = where(a eq a and flags eq flags,wc)
            if (wc gt 0) then y2[i-lo2] = median(a[w]) else y2[i-lo2] = -999.
        endfor

        ; set total arrays for combined above and below row numbers and medians.
        x = [x1,x2]
        y = [y1,y2]

        w = where(y gt (-10.),wc) ; i.e., omit -999. values.

        if (wc gt min_rows) then begin
            ; require minimum rows with data for interpolation
            x = x[w]
            y = y[w]
            b = poly_fit(double(x),double(y),1)
            ; interpolate background level for each row in transition region,
            ; and force the row to that level.
            for i = n-hw_fix,n+hw_fix do begin
                a_all = im[cc[0]:cc[1],i]
                flags_all = rowflag[cc[0]:cc[1]]
                w_all = where(a_all eq a_all,wc_all)
                ; correction is calculated only over low-median columns
                a = a_all[ptr_col]
                flags = flags_all[ptr_col]
                w = where(a eq a,wc)
                w2 = where(a eq a and flags eq flags,w2c)
                if (w2c gt 0) then begin
                    y_im = median(a[w2])
                    y_want = poly(float(i),b)
                    del = y_want - y_im
                    ; apply correction to all columns in partition
                    a_all[w_all] = a_all[w_all] + del
                    imout[cc[0]:cc[1],i] = a_all
                    ; a[w] = a[w] + del ; adjust all non-NaNs.
                    ; imout[cc[0]:cc[1],i] = a
                endif ; sufficient pixels to calculate median
            endfor ; loop over transition rows to be fixed.
        endif ; sufficient rows to do interpolation

    endif ; valid column partition

endfor ; loop over column partitions

end
    
